"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/modules/token_embedders/pretrained_transformer_mismatched_embedder.py
"""

from typing import Optional, Dict, Any


import torch

from combo.config.from_parameters import register_arguments
from combo.modules.token_embedders import PretrainedTransformerEmbedder, TokenEmbedder

from combo.config import Registry
from combo.nn.utils import batched_span_select
from combo.utils import ConfigurationError


@Registry.register( "pretrained_transformer_mismatched_token_embedder")
class PretrainedTransformerMismatchedEmbedder(TokenEmbedder):
    """
    Use this embedder to embed wordpieces given by `PretrainedTransformerMismatchedIndexer`
    and to get word-level representations.

    Registered as a `TokenEmbedder` with name "pretrained_transformer_mismatched".

    # Parameters

    model_name : `str`
        The name of the `transformers` model to use. Should be the same as the corresponding
        `PretrainedTransformerMismatchedIndexer`.
    max_length : `int`, optional (default = `None`)
        If positive, folds input token IDs into multiple segments of this length, pass them
        through the transformer model independently, and concatenate the final representations.
        Should be set to the same value as the `max_length` option on the
        `PretrainedTransformerMismatchedIndexer`.
    sub_module: `str`, optional (default = `None`)
        The name of a submodule of the transformer to be used as the embedder. Some transformers naturally act
        as embedders such as BERT. However, other models consist of encoder and decoder, in which case we just
        want to use the encoder.
    train_parameters: `bool`, optional (default = `True`)
        If this is `True`, the transformer weights get updated during training.
    last_layer_only: `bool`, optional (default = `True`)
        When `True` (the default), only the final layer of the pretrained transformer is taken
        for the embeddings. But if set to `False`, a scalar mix of all of the layers
        is used.
    override_weights_file: `Optional[str]`, optional (default = `None`)
        If set, this specifies a file from which to load alternate weights that override the
        weights from huggingface. The file is expected to contain a PyTorch `state_dict`, created
        with `torch.save()`.
    override_weights_strip_prefix: `Optional[str]`, optional (default = `None`)
        If set, strip the given prefix from the state dict when loading it.
    load_weights: `bool`, optional (default = `True`)
        Whether to load the pretrained weights. If you're loading your model/predictor from an AllenNLP archive
        it usually makes sense to set this to `False` (via the `overrides` parameter)
        to avoid unnecessarily caching and loading the original pretrained weights,
        since the archive will already contain all of the weights needed.
    gradient_checkpointing: `bool`, optional (default = `None`)
        Enable or disable gradient checkpointing.
    tokenizer_kwargs: `Dict[str, Any]`, optional (default = `None`)
        Dictionary with
        [additional arguments](https://github.com/huggingface/transformers/blob/155c782a2ccd103cf63ad48a2becd7c76a7d2115/transformers/tokenization_utils.py#L691)
        for `AutoTokenizer.from_pretrained`.
    transformer_kwargs: `Dict[str, Any]`, optional (default = `None`)
        Dictionary with
        [additional arguments](https://github.com/huggingface/transformers/blob/155c782a2ccd103cf63ad48a2becd7c76a7d2115/transformers/modeling_utils.py#L253)
        for `AutoModel.from_pretrained`.
    sub_token_mode: `Optional[str]`, optional (default= `avg`)
        If `sub_token_mode` is set to `first`, return first sub-token representation as word-level representation
        If `sub_token_mode` is set to `avg`, return average of all the sub-tokens representation as word-level representation
        If `sub_token_mode` is not specified it defaults to `avg`
        If invalid `sub_token_mode` is provided, throw `ConfigurationError`

    """  # noqa: E501
    @register_arguments
    def __init__(
        self,
        model_name: str,
        max_length: int = None,
        sub_module: str = None,
        train_parameters: bool = True,
        last_layer_only: bool = True,
        override_weights_file: Optional[str] = None,
        override_weights_strip_prefix: Optional[str] = None,
        load_weights: bool = True,
        gradient_checkpointing: Optional[bool] = None,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        transformer_kwargs: Optional[Dict[str, Any]] = None,
        sub_token_mode: Optional[str] = "avg",
    ) -> None:
        super().__init__()
        # The matched version v.s. mismatched
        self._matched_embedder = PretrainedTransformerEmbedder(
            model_name,
            max_length=max_length,
            sub_module=sub_module,
            train_parameters=train_parameters,
            last_layer_only=last_layer_only,
            override_weights_file=override_weights_file,
            override_weights_strip_prefix=override_weights_strip_prefix,
            load_weights=load_weights,
            gradient_checkpointing=gradient_checkpointing,
            tokenizer_kwargs=tokenizer_kwargs,
            transformer_kwargs=transformer_kwargs,
        )
        self.sub_token_mode = sub_token_mode

    def get_output_dim(self):
        return self._matched_embedder.get_output_dim()

    def forward(
        self,
        token_ids: torch.LongTensor,
        mask: torch.BoolTensor,
        offsets: torch.LongTensor,
        wordpiece_mask: torch.BoolTensor,
        type_ids: Optional[torch.LongTensor] = None,
        segment_concat_mask: Optional[torch.BoolTensor] = None,
    ) -> torch.Tensor:  # type: ignore
        """
        # Parameters

        token_ids: `torch.LongTensor`
            Shape: [batch_size, num_wordpieces] (for exception see `PretrainedTransformerEmbedder`).
        mask: `torch.BoolTensor`
            Shape: [batch_size, num_orig_tokens].
        offsets: `torch.LongTensor`
            Shape: [batch_size, num_orig_tokens, 2].
            Maps indices for the original tokens, i.e. those given as input to the indexer,
            to a span in token_ids. `token_ids[i][offsets[i][j][0]:offsets[i][j][1] + 1]`
            corresponds to the original j-th token from the i-th batch.
        wordpiece_mask: `torch.BoolTensor`
            Shape: [batch_size, num_wordpieces].
        type_ids: `Optional[torch.LongTensor]`
            Shape: [batch_size, num_wordpieces].
        segment_concat_mask: `Optional[torch.BoolTensor]`
            See `PretrainedTransformerEmbedder`.

        # Returns

        `torch.Tensor`
            Shape: [batch_size, num_orig_tokens, embedding_size].
        """
        # Shape: [batch_size, num_wordpieces, embedding_size].
        embeddings = self._matched_embedder(
            token_ids, wordpiece_mask, type_ids=type_ids, segment_concat_mask=segment_concat_mask
        )

        # span_embeddings: (batch_size, num_orig_tokens, max_span_length, embedding_size)
        # span_mask: (batch_size, num_orig_tokens, max_span_length)
        span_embeddings, span_mask = batched_span_select(embeddings.contiguous(), offsets)

        span_mask = span_mask.unsqueeze(-1)

        # Shape: (batch_size, num_orig_tokens, max_span_length, embedding_size)
        span_embeddings *= span_mask  # zero out paddings

        # If "sub_token_mode" is set to "first", return the first sub-token embedding
        if self.sub_token_mode == "first":
            # Select first sub-token embeddings from span embeddings
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            orig_embeddings = span_embeddings[:, :, 0, :]

        # If "sub_token_mode" is set to "avg", return the average of embeddings of all sub-tokens of a word
        elif self.sub_token_mode == "avg":
            # Sum over embeddings of all sub-tokens of a word
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            span_embeddings_sum = span_embeddings.sum(2)

            # Shape (batch_size, num_orig_tokens)
            span_embeddings_len = span_mask.sum(2)

            # Find the average of sub-tokens embeddings by dividing `span_embedding_sum` by `span_embedding_len`
            # Shape: (batch_size, num_orig_tokens, embedding_size)
            orig_embeddings = span_embeddings_sum / torch.clamp_min(span_embeddings_len, 1)

            # All the places where the span length is zero, write in zeros.
            orig_embeddings[(span_embeddings_len == 0).expand(orig_embeddings.shape)] = 0

        # If invalid "sub_token_mode" is provided, throw error
        else:
            raise ConfigurationError(f"Do not recognise 'sub_token_mode' {self.sub_token_mode}")

        return orig_embeddings



from typing import List

from combo.config import Registry
from combo.data.tokenizers.token import Token
from combo.data.tokenizers.tokenizer import Tokenizer


@Registry.register('whitespace_tokenizer')
class WhitespaceTokenizer(Tokenizer):
    """
    A `Tokenizer` that assumes you've already done your own tokenization somehow and have
    separated the tokens by spaces.  We just split the input string on whitespace and return the
    resulting list.
    Note that we use `text.split()`, which means that the amount of whitespace between the
    tokens does not matter.  This will never result in spaces being included as tokens.
    Registered as a `Tokenizer` with name "whitespace" and "just_spaces".
    """

    def tokenize(self, text: str) -> List[Token]:
        return [Token(t) for t in text.split()]



import unittest

from combo.data import LamboTokenizer


class LamboTokenizerTest(unittest.TestCase):

    def setUp(self) -> None:
        self.lambo_tokenizer = LamboTokenizer()

    def test_tokenize_sentence(self):
        tokens = self.lambo_tokenizer.tokenize('Hello cats. I love you')
        self.assertListEqual([t.text for t in tokens],
                             ['Hello', 'cats', '.', 'I', 'love', 'you'])

    def test_segment_text(self):
        tokens = self.lambo_tokenizer.segment('Hello cats. I love you.\n\nHi.')
        self.assertListEqual(tokens,
                             [['Hello', 'cats', '.'], ['I', 'love', 'you', '.'], ['Hi', '.']])

    def test_segment_text_with_turns(self):
        tokens = self.lambo_tokenizer.segment('Hello cats. I love you.\n\nHi.', turns=True)
        self.assertListEqual(tokens,
                             [['Hello', 'cats', '.', 'I', 'love', 'you', '.'], ['Hi', '.']])

    def test_segment_text_with_multiwords(self):
        tokens = self.lambo_tokenizer.segment('I don\'t want a pizza.', split_subwords=True)
        self.assertListEqual(tokens,
                             [['I', 'do', 'n\'t', 'want', 'a', 'pizza', '.']])

    def test_segment_text_with_multiwords_without_splitting(self):
        tokens = self.lambo_tokenizer.segment('I don\'t want a pizza.', split_subwords=False)
        self.assertListEqual(tokens,
                             [['I', 'don\'t', 'want', 'a', 'pizza', '.']])

    def test_tokenize_sentence_with_multiword(self):
        tokens = self.lambo_tokenizer.tokenize('I don\'t like apples.')
        self.assertListEqual([t.text for t in tokens],
                             ['I', 'don\'t', 'like', 'apples', '.'])
        self.assertListEqual([t.subwords for t in tokens],
                             [[], ['do', 'n\'t'], [], [], []])



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/data/fields/metadata_field.py
"""
from typing import Any, Dict, List, Mapping


from combo.data.fields.field import DataArray, Field


class MetadataField(Field[DataArray], Mapping[str, Any]):
    """
    A `MetadataField` is a `Field` that does not get converted into tensors.  It just carries
    side information that might be needed later on, for computing some third-party metric, or
    outputting debugging information, or whatever else you need.  We use this in the BiDAF model,
    for instance, to keep track of question IDs and passage token offsets, so we can more easily
    use the official evaluation script to compute metrics.
    We don't try to do any kind of smart combination of this field for batched input - when you use
    this `Field` in a model, you'll get a list of metadata objects, one for each instance in the
    batch.
    # Parameters
    metadata : `Any`
        Some object containing the metadata that you want to store.  It's likely that you'll want
        this to be a dictionary, but it could be anything you want.
    """

    __slots__ = ["metadata"]

    def __init__(self, metadata: Any) -> None:
        self.metadata = metadata

    def __getitem__(self, key: str) -> Any:
        try:
            return self.metadata[key]  # type: ignore
        except TypeError:
            raise TypeError("your metadata is not a dict")

    def __iter__(self):
        try:
            return iter(self.metadata)
        except TypeError:
            raise TypeError("your metadata is not iterable")

    def __len__(self):
        try:
            return len(self.metadata)
        except TypeError:
            raise TypeError("your metadata has no length")

    def get_padding_lengths(self) -> Dict[str, int]:
        return {}

    def as_tensor(self, padding_lengths: Dict[str, int]) -> DataArray:

        return self.metadata  # type: ignore

    def empty_field(self) -> "MetadataField":
        return MetadataField(None)

    def batch_tensors(self, tensor_list: List[DataArray]) -> List[DataArray]:  # type: ignore
        return tensor_list

    def __str__(self) -> str:
        return "MetadataField (print field.metadata to see specific information)."

    def human_readable_repr(self):
        if hasattr(self.metadata, "human_readable_repr"):
            return self.metadata.human_readable_repr()
        return self.metadata



"""
Adapted from COMBO
Author: Mateusz Klimaszewski
"""
from typing import Any, Dict, Optional

import torch
from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.modules.token_embedders.pretrained_transformer_mismatched_embedder import \
    PretrainedTransformerMismatchedEmbedder
from combo.nn import base
from combo.nn.activations import Activation


@Registry.register("transformers_word_embedder")
class TransformersWordEmbedder(PretrainedTransformerMismatchedEmbedder):
    """
    Transformers word embeddings as last hidden state + optional projection layers.

    Tested with Bert (but should work for other models as well).
    """

    authorized_missing_keys = [r"position_ids$"]

    @register_arguments
    def __init__(self,
                 model_name: str,
                 projection_dim: int = 0,
                 projection_activation: Optional[Activation] = lambda x: x,
                 projection_dropout_rate: Optional[float] = 0.0,
                 freeze_transformer: bool = True,
                 last_layer_only: bool = True,
                 tokenizer_kwargs: Optional[Dict[str, Any]] = None,
                 transformer_kwargs: Optional[Dict[str, Any]] = None):
        super().__init__(model_name,
                         train_parameters=not freeze_transformer,
                         last_layer_only=last_layer_only,
                         tokenizer_kwargs=tokenizer_kwargs,
                         transformer_kwargs=transformer_kwargs)
        if projection_dim:
            self.projection_layer = base.Linear(in_features=super().get_output_dim(),
                                                out_features=projection_dim,
                                                dropout_rate=projection_dropout_rate,
                                                activation=projection_activation)
            self.output_dim = projection_dim
        else:
            self.projection_layer = None
            self.output_dim = super().get_output_dim()

    #@overrides
    def forward(
            self,
            token_ids: torch.LongTensor,
            mask: torch.BoolTensor,
            offsets: torch.LongTensor,
            wordpiece_mask: torch.BoolTensor,
            type_ids: Optional[torch.LongTensor] = None,
            segment_concat_mask: Optional[torch.BoolTensor] = None,
    ) -> torch.Tensor:  # type: ignore
        x = super().forward(token_ids, mask, offsets, wordpiece_mask, type_ids, segment_concat_mask)
        if self.projection_layer:
            x = self.projection_layer(x)
        return x

    @overrides
    def get_output_dim(self) -> int:
        return self.output_dim



import logging
from overrides import overrides
from datetime import datetime


class ComboLogger(logging.Logger):
    def __init__(self, name: str, prefix: str = None, display_date: bool = True):
        super().__init__(name)
        self.__prefix = prefix or ''
        self.__display_date = display_date

    @overrides(check_signature=False)
    def log(self, level: int, msg: str, prefix: str = None):
        prefix = prefix or self.__prefix
        super().log(level, '[{date} UTC {prefix}] {msg}'.format(
            date=datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'),
            prefix=prefix,
            msg=msg
        ))

    @overrides(check_signature=False)
    def debug(self, msg: str, prefix: str = None):
        self.log(level=logging.DEBUG, msg=msg, prefix=prefix)

    @overrides(check_signature=False)
    def info(self, msg: str, prefix: str = None):
        self.log(level=logging.INFO, msg=msg, prefix=prefix)

    @overrides(check_signature=False)
    def warning(self, msg: str, prefix: str = None):
        self.log(level=logging.WARN, msg=msg, prefix=prefix)

    @overrides(check_signature=False)
    def error(self, msg: str, prefix: str = None):
        self.log(level=logging.ERROR, msg=msg, prefix=prefix)

    @overrides(check_signature=False)
    def fatal(self, msg: str, prefix: str = None):
        self.log(level=logging.FATAL, msg=msg, prefix=prefix)

    @overrides(check_signature=False)
    def critical(self, msg: str, prefix: str = None):
        self.log(level=logging.CRITICAL, msg=msg, prefix=prefix)



from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data import Vocabulary
from combo.nn.utils import masked_cross_entropy
from combo.nn.base import FeedForward
from combo.predictors import Predictor
import torch
from typing import Dict, List, Optional, Union

from combo.utils import ConfigurationError


@Registry.register("feedforward_predictor")
@Registry.register("feedforward_predictor_from_vocab", constructor_method="from_vocab")
class FeedForwardPredictor(Predictor):
    """Feedforward predictor. Should be used on top of Seq2Seq encoder."""

    @register_arguments
    def __init__(self, feedforward_network: "FeedForward"):
        super().__init__()
        self.feedforward_network = feedforward_network

    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        if mask is None:
            mask = x.new_ones(x.size()[:-1])

        x, feature_maps = self.feedforward_network(x)
        output = {
            "prediction": x.argmax(-1),
            "probability": x,
            "embedding": feature_maps[-1],
        }

        if labels is not None:
            if sample_weights is None:
                sample_weights = labels.new_ones([mask.size(0)])
            output["loss"] = self._loss(x, labels, mask, sample_weights)

        return output

    def _loss(self,
              pred: torch.Tensor,
              true: torch.Tensor,
              mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> torch.Tensor:
        BATCH_SIZE, _, CLASSES = pred.size()
        valid_positions = mask.sum()
        pred = pred.reshape(-1, CLASSES)
        true = true.reshape(-1)
        mask = mask.reshape(-1)
        loss = masked_cross_entropy(pred, true, mask)
        loss = loss.reshape(BATCH_SIZE, -1) * sample_weights.unsqueeze(-1)
        return loss.sum() / valid_positions

    @classmethod
    @register_arguments
    def from_vocab(cls,
                   vocabulary: Vocabulary,
                   vocab_namespace: str,
                   input_dim: int,
                   num_layers: int,
                   hidden_dims: List[int],
                   activations: Union["Activation", List["Activation"]],
                   dropout: Union[float, List[float]] = 0.0,
                   ):
        if len(hidden_dims) + 1 != num_layers:
            raise ConfigurationError(
                f"len(hidden_dims) ({len(hidden_dims):d}) + 1 != num_layers ({num_layers:d})"
            )

        assert vocab_namespace in vocabulary.get_namespaces(), \
            f"There is not {vocab_namespace} in created vocabs, check if this field has any values to predict!"
        hidden_dims = hidden_dims + [vocabulary.get_vocab_size(vocab_namespace)]

        ff_p = cls(FeedForward(
            input_dim=input_dim,
            num_layers=num_layers,
            hidden_dims=hidden_dims,
            activations=activations,
            dropout=dropout))

        ff_p.constructed_from = "from_vocab"
        return ff_p


import unittest
import os

from combo.data.tokenizers import Token
from combo.data.token_indexers import TokenCharactersIndexer
from combo.data.vocabulary import Vocabulary


class TokenCharactersIndexerTest(unittest.TestCase):

    def setUp(self):
        self.vocabulary = Vocabulary.from_files(
            os.path.normpath(os.path.join(os.getcwd(), '../../fixtures/train_vocabulary')),
            oov_token='_',
            padding_token='__PAD__'
        )
        self.token_characters_indexer = TokenCharactersIndexer()

    def test_get_index_to_token(self):
        token = Token(idx=0, text='a')
        counter = {'token_characters': {'a': 0}}
        self.token_characters_indexer.count_vocab_items(token, counter)
        self.assertEqual(counter['token_characters']['a'], 1)

    def test_tokens_to_indices(self):
        self.assertEqual(
            self.token_characters_indexer.tokens_to_indices(
                [Token('za')], self.vocabulary),
            {'token_characters': [[8, 4]]})

    def test_get_padding_lengths(self):
        self.assertEqual(
            self.token_characters_indexer.get_padding_lengths({'token_characters': [[11, 4]]}),
            {'num_token_characters': 2, 'token_characters': 1}
        )

    def test_get_nondefault_padding_lengths(self):
        token_characters_indexer_w_padding = TokenCharactersIndexer(min_padding_length=4)
        self.assertEqual(
            token_characters_indexer_w_padding.get_padding_lengths({'token_characters': [[11, 4]]}),
            {'num_token_characters': 4, 'token_characters': 1}
        )

    def test_tokens_to_indices_as_padded_tensor(self):
        token_characters_indexer_w_padding = TokenCharactersIndexer(min_padding_length=4)
        indices_dict = token_characters_indexer_w_padding.tokens_to_indices([Token('wa')], self.vocabulary)
        padded_dict = list(token_characters_indexer_w_padding.as_padded_tensor_dict(
            indices_dict,
            token_characters_indexer_w_padding.get_padding_lengths(indices_dict)
        )['token_characters'].numpy()[0])
        self.assertListEqual(padded_dict,[11, 4, 0, 0])



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/common/util.py#L261
"""
import logging
from typing import Dict, Tuple

import spacy
from spacy.cli.download import download as spacy_download
from spacy.language import Language as SpacyModelType

logger = logging.getLogger(__name__)

LOADED_SPACY_MODELS: Dict[Tuple[str, bool, bool, bool], SpacyModelType] = {}


def get_spacy_model(
    spacy_model_name: str, pos_tags: bool = True, parse: bool = False, ner: bool = False
) -> SpacyModelType:
    """
    In order to avoid loading spacy models a whole bunch of times, we'll save references to them,
    keyed by the options we used to create the spacy model, so any particular configuration only
    gets loaded once.
    """

    options = (spacy_model_name, pos_tags, parse, ner)
    if options not in LOADED_SPACY_MODELS:
        disable = ["vectors", "textcat"]
        if not pos_tags:
            disable.append("tagger")
        if not parse:
            disable.append("parser")
        if not ner:
            disable.append("ner")
        try:
            spacy_model = spacy.load(spacy_model_name, disable=disable)
        except OSError:
            logger.warning(
                f"Spacy models '{spacy_model_name}' not found.  Downloading and installing."
            )
            spacy_download(spacy_model_name)

            # Import the downloaded model module directly and load from there
            spacy_model_module = __import__(spacy_model_name)
            spacy_model = spacy_model_module.load(disable=disable)  # type: ignore

        LOADED_SPACY_MODELS[options] = spacy_model
    return LOADED_SPACY_MODELS[options]


import os
import unittest
from tempfile import TemporaryDirectory

from combo.combo_model import ComboModel
from combo.data.vocabulary import Vocabulary
from combo.default_model import default_model
from combo.modules import archive

TEMP_FILE_PATH = 'temp_serialization_dir'


def _test_vocabulary() -> Vocabulary:
    return Vocabulary.from_files(os.path.normpath(os.path.join(os.getcwd(), '../fixtures/train_vocabulary')),
                                 oov_token='_', padding_token='__PAD__')


class ArchivalTest(unittest.TestCase):
    def test_serialize_model(self):
        vocabulary = _test_vocabulary()
        model = default_model('bert-base-cased', vocabulary)
        t = '.'
        with TemporaryDirectory(TEMP_FILE_PATH) as t:
            archive(model, t)
            loaded_model = ComboModel.from_archive(os.path.join(t, 'model.tar.gz'))
        self.assertDictEqual(loaded_model.serialize(), model.serialize())



"""Features indexer.
Adapted from COMBO.
Author: Mateusz Klimaszewski
"""
import collections
from typing import List, Dict

import torch
from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.tokenizers.tokenizer import Token
from combo.data.token_indexers.token_indexer import TokenIndexer, IndexedTokenList
from combo.utils import pad_sequence_to_length


@Registry.register('feats_indexer')
class TokenFeatsIndexer(TokenIndexer):
    @register_arguments
    def __init__(
            self,
            namespace: str = "feats_labels",
            feature_name: str = "feats",
            token_min_padding_length: int = 0,
    ) -> None:
        super().__init__(token_min_padding_length)
        self.namespace = namespace
        self._feature_name = feature_name

    @overrides
    def count_vocab_items(self, token: Token, counter: Dict[str, Dict[str, int]]):
        feats = self._feat_values(token)
        for feat in feats:
            counter[self.namespace][feat] += 1

    @overrides
    def tokens_to_indices(self, tokens: List[Token], vocabulary) -> IndexedTokenList:
        indices: List[List[int]] = []
        vocab_size = vocabulary.get_vocab_size(self.namespace)
        for token in tokens:
            token_indices = []
            feats = self._feat_values(token)
            for feat in feats:
                token_indices.append(vocabulary.get_token_index(feat, self.namespace))
            indices.append(pad_sequence_to_length(token_indices, vocab_size))
        return {"tokens": indices}

    @overrides
    def get_empty_token_list(self) -> IndexedTokenList:
        return {"tokens": [[]]}

    def _feat_values(self, token):
        feats = getattr(token, self._feature_name)
        if feats is None:
            feats = collections.OrderedDict()
        features = []
        for feat, value in feats.items():
            if feat in ["_", "__ROOT__"]:
                pass
            else:
                # Handle case where feature is binary (doesn't have associated value)
                if value:
                    features.append(feat + "=" + value)
                else:
                    features.append(feat)
        return features

    @overrides
    def as_padded_tensor_dict(
            self, tokens: IndexedTokenList, padding_lengths: Dict[str, int]
    ) -> Dict[str, torch.Tensor]:
        tensor_dict = {}
        for key, val in tokens.items():
            vocab_size = len(val[0])
            tensor = torch.tensor(pad_sequence_to_length(val,
                                                         padding_lengths[key],
                                                         default_value=lambda: [0] * vocab_size,
                                                         )
                                  )
            tensor_dict[key] = tensor
        return tensor_dict




import torch.distributed as dist


def is_distributed() -> bool:
    """
    Checks if the distributed process group is available and has been initialized
    """
    return dist.is_available() and dist.is_initialized()


class MalformedFileException(Exception):
    def __init__(self, message):
        super().__init__(message)



from typing import Union, Iterable, TypeVar, List, Iterator, Any
from itertools import islice

import numpy
import spacy
import torch
import hashlib
import io
import base58
import dill

A = TypeVar("A")


def hash_object(o: Any) -> str:
    """Adapted from AllenNLP"""
    """Returns a character hash code of arbitrary Python objects."""
    m = hashlib.blake2b()
    with io.BytesIO() as buffer:
        dill.dump(o, buffer)
        m.update(buffer.getbuffer())
        return base58.b58encode(m.digest()).decode()


def int_to_device(device: Union[int, torch.device]) -> torch.device:
    if isinstance(device, torch.device):
        return device
    if device < 0:
        return torch.device("cpu")
    return torch.device(device)


def ensure_list(iterable: Iterable[A]) -> List[A]:
    """
    An Iterable may be a list or a generator.
    This ensures we get a list without making an unnecessary copy.
    """
    if isinstance(iterable, list):
        return iterable
    else:
        return list(iterable)


def lazy_groups_of(iterable: Iterable[A], group_size: int) -> Iterator[List[A]]:
    """
    Takes an iterable and batches the individual instances into lists of the
    specified size. The last list may be smaller if there are instances left over.
    """
    iterator = iter(iterable)
    while True:
        s = list(islice(iterator, group_size))
        if len(s) > 0:
            yield s
        else:
            break


def sanitize(x: Any) -> Any:
    """
    Sanitize turns PyTorch and Numpy types into basic Python types so they
    can be serialized into JSON.
    """
    # Import here to avoid circular references
    from combo.data.tokenizers import Token

    if isinstance(x, (str, float, int, bool)):
        # x is already serializable
        return x
    elif isinstance(x, torch.Tensor):
        # tensor needs to be converted to a list (and moved to cpu if necessary)
        return x.cpu().tolist()
    elif isinstance(x, numpy.ndarray):
        # array needs to be converted to a list
        return x.tolist()
    elif isinstance(x, numpy.number):
        # NumPy numbers need to be converted to Python numbers
        return x.item()
    elif isinstance(x, dict):
        # Dicts need their values sanitized
        return {key: sanitize(value) for key, value in x.items()}
    elif isinstance(x, numpy.bool_):
        # Numpy bool_ need to be converted to python bool.
        return bool(x)
    elif isinstance(x, (spacy.tokens.Token, Token)):
        # Tokens get sanitized to just their text.
        return x.text
    elif isinstance(x, (list, tuple, set)):
        # Lists, tuples, and sets need their values sanitized
        return [sanitize(x_i) for x_i in x]
    elif x is None:
        return "None"
    elif hasattr(x, "to_json"):
        return x.to_json()
    else:
        raise ValueError(
            f"Cannot sanitize {x} of type {type(x)}. "
            "If this is your own custom class, add a `to_json(self)` method "
            "that returns a JSON-like object."
        )



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/common/util.py
"""

from typing import Any, Callable, List, Sequence


def sanitize_wordpiece(wordpiece: str) -> str:
    """
    Sanitizes wordpieces from BERT, RoBERTa or ALBERT tokenizers.
    """
    if wordpiece.startswith("##"):
        return wordpiece[2:]
    elif wordpiece.startswith("Ġ"):
        return wordpiece[1:]
    elif wordpiece.startswith("▁"):
        return wordpiece[1:]
    else:
        return wordpiece


def pad_sequence_to_length(
    sequence: Sequence,
    desired_length: int,
    default_value: Callable[[], Any] = lambda: 0,
    padding_on_right: bool = True,
) -> List:
    """
    Take a list of objects and pads it to the desired length, returning the padded list.  The
    original list is not modified.
    # Parameters
    sequence : `List`
        A list of objects to be padded.
    desired_length : `int`
        Maximum length of each sequence. Longer sequences are truncated to this length, and
        shorter ones are padded to it.
    default_value: `Callable`, optional (default=`lambda: 0`)
        Callable that outputs a default value (of any type) to use as padding values.  This is
        a lambda to avoid using the same object when the default value is more complex, like a
        list.
    padding_on_right : `bool`, optional (default=`True`)
        When we add padding tokens (or truncate the sequence), should we do it on the right or
        the left?
    # Returns
    padded_sequence : `List`
    """
    sequence = list(sequence)
    # Truncates the sequence to the desired length.
    if padding_on_right:
        padded_sequence = sequence[:desired_length]
    else:
        padded_sequence = sequence[-desired_length:]
    # Continues to pad with default_value() until we reach the desired length.
    pad_length = desired_length - len(padded_sequence)
    # This just creates the default value once, so if it's a list, and if it gets mutated
    # later, it could cause subtle bugs. But the risk there is low, and this is much faster.
    values_to_pad = [default_value()] * pad_length
    if padding_on_right:
        padded_sequence = padded_sequence + values_to_pad
    else:
        padded_sequence = values_to_pad + padded_sequence
    return padded_sequence



"""
Re-implemented AllenNLP BatchSampler
https://github.com/allenai/allennlp/blob/main/allennlp/data/samplers/batch_sampler.py
"""
from typing import Sequence, Iterable, List, Optional
from torch import Tensor

from combo.config import Registry, FromParameters


@Registry.register('base_batch_sampler')
class BatchSampler(FromParameters):
    def get_batch_indices(self, instances: Sequence[Tensor]) -> Iterable[List[int]]:
        raise NotImplementedError

    def get_num_batches(self, instances: Sequence[Tensor]) -> int:
        raise NotImplementedError

    def get_batch_size(self) -> Optional[int]:
        """
        Not all `BatchSamplers` define a consistent `batch_size`, but those that
        do should override this method.
        """
        return None



from .adjacency_field import AdjacencyField
from .field import Field
from .label_field import LabelField
from .list_field import ListField
from .metadata_field import MetadataField
from .sequence_field import SequenceField
from .sequence_label_field import SequenceLabelField
from .sequence_multilabel_field import SequenceMultiLabelField
from .text_field import TextField



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/nn/module.py#L14
"""

from typing import List, Tuple
import torch

from combo.nn.utils import (
    _check_incompatible_keys,
    _IncompatibleKeys,
    StateDictType,
)


class Module(torch.nn.Module):
    """
    This is just `torch.nn.Module` with some extra functionality.
    """

    def _post_load_state_dict(
        self, missing_keys: List[str], unexpected_keys: List[str]
    ) -> Tuple[List[str], List[str]]:
        """
        Subclasses can override this and potentially modify `missing_keys` or `unexpected_keys`.
        """
        return missing_keys, unexpected_keys

    def load_state_dict(self, state_dict: StateDictType, strict: bool = True) -> _IncompatibleKeys:
        """
        Same as [`torch.nn.Module.load_state_dict()`]
        (https://pytorch.org/docs/master/generated/torch.nn.Module.html#torch.nn.Module.load_state_dict)
        except we also run the [`_post_load_state_dict`](#_post_load_state_dict) method before returning,
        which can be implemented by subclasses to customize the behavior.
        """
        missing_keys, unexpected_keys = super().load_state_dict(state_dict, strict=False)  # type: ignore[arg-type]
        missing_keys, unexpected_keys = self._post_load_state_dict(missing_keys, unexpected_keys)
        _check_incompatible_keys(self, missing_keys, unexpected_keys, strict)
        return _IncompatibleKeys(missing_keys, unexpected_keys)



from .token_indexer import IndexedTokenList, TokenIndexer
from .token_features_indexer import TokenFeatsIndexer
from .single_id_token_indexer import SingleIdTokenIndexer
from .pretrained_transformer_indexer import PretrainedTransformerIndexer
from .pretrained_transformer_mismatched_indexer import PretrainedTransformerMismatchedIndexer
from .pretrained_transformer_fixed_mismatched_indexer import PretrainedTransformerFixedMismatchedIndexer
from .token_const_padding_characters_indexer import TokenConstPaddingCharactersIndexer
from .token_characters_indexer import TokenCharactersIndexer



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/data/fields/label_field.py
"""

from typing import Dict, Union, Set
import logging


import torch

from combo.data.vocabulary import Vocabulary
from combo.data.fields import Field
from combo.utils import ConfigurationError

logger = logging.getLogger(__name__)


class LabelField(Field[torch.Tensor]):
    """
    A `LabelField` is a categorical label of some kind, where the labels are either strings of
    text or 0-indexed integers (if you wish to skip indexing by passing skip_indexing=True).
    If the labels need indexing, we will use a :class:`Vocabulary` to convert the string labels
    into integers.
    This field will get converted into an integer index representing the class label.
    # Parameters
    label : `Union[str, int]`
    label_namespace : `str`, optional (default=`"labels"`)
        The _namespace to use for converting label strings into integers.  We map label strings to
        integers for you (e.g., "entailment" and "contradiction" get converted to 0, 1, ...),
        and this _namespace tells the `Vocabulary` object which mapping from strings to integers
        to use (so "entailment" as a label doesn't get the same integer idx as "entailment" as a
        word).  If you have multiple different label fields in your data, you should make sure you
        use different namespaces for each one, always using the suffix "labels" (e.g.,
        "passage_labels" and "question_labels").
    skip_indexing : `bool`, optional (default=`False`)
        If your labels are 0-indexed integers, you can pass in this flag, and we'll skip the indexing
        step.  If this is `False` and your labels are not strings, this throws a `ConfigurationError`.
    """

    __slots__ = ["label", "_label_namespace", "_label_id", "_skip_indexing"]

    # Most often, you probably don't want to have OOV/PAD tokens with a LabelField, so we warn you
    # about it when you pick a _namespace that will getting these tokens by default.  It is
    # possible, however, that you _do_ actually want OOV/PAD tokens with this Field.  This class
    # variable is used to make sure that we only log a single warning for this per _namespace, and
    # not every time you create one of these Field objects.
    _already_warned_namespaces: Set[str] = set()

    def __init__(
        self, label: Union[str, int], label_namespace: str = "labels", skip_indexing: bool = False
    ) -> None:
        self.label = label
        self._label_namespace = label_namespace
        self._label_id = None
        self._maybe_warn_for_namespace(label_namespace)
        self._skip_indexing = skip_indexing

        if skip_indexing:
            if not isinstance(label, int):
                raise ConfigurationError(
                    "In order to skip indexing, your labels must be integers. "
                    "Found label = {}".format(label)
                )
            self._label_id = label
        elif not isinstance(label, str):
            raise ConfigurationError(
                "LabelFields must be passed a string label if skip_indexing=False. "
                "Found label: {} with type: {}.".format(label, type(label))
            )

    def _maybe_warn_for_namespace(self, label_namespace: str) -> None:
        if not (self._label_namespace.endswith("labels") or self._label_namespace.endswith("tags")):
            if label_namespace not in self._already_warned_namespaces:
                logger.warning(
                    "Your label _namespace was '%s'. We recommend you use a _namespace "
                    "ending with 'labels' or 'tags', so we don't add UNK and PAD tokens by "
                    "default to your vocabulary.  See documentation for "
                    "`non_padded_namespaces` parameter in Vocabulary.",
                    self._label_namespace,
                )
                self._already_warned_namespaces.add(label_namespace)

    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]):
        if self._label_id is None:
            counter[self._label_namespace][self.label] += 1  # type: ignore

    def index(self, vocab: Vocabulary):
        if not self._skip_indexing:
            self._label_id = vocab.get_token_index(
                self.label, self._label_namespace  # type: ignore
            )

    def get_padding_lengths(self) -> Dict[str, int]:
        return {}

    def as_tensor(self, padding_lengths: Dict[str, int]) -> torch.Tensor:
        tensor = torch.tensor(self._label_id, dtype=torch.long)
        return tensor

    def empty_field(self):
        return LabelField(-1, self._label_namespace, skip_indexing=True)

    def human_readable_repr(self) -> Union[str, int]:
        return self.label

    def __str__(self) -> str:
        return f"LabelField with label: {self.label} in _namespace: '{self._label_namespace}'."

    def __len__(self):
        return 1



import os
import unittest

from combo.data import SingleIdTokenIndexer, Token
from combo.data.vocabulary import Vocabulary


class VocabularyTest(unittest.TestCase):

    def setUp(self):
        self.vocabulary = Vocabulary.from_files(
            os.path.join(os.getcwd(), '../../fixtures/train_vocabulary'),
            oov_token='_',
            padding_token='__PAD__'
        )
        self.single_id_indexer = SingleIdTokenIndexer(namespace='token_characters')

    def test_get_index_to_token(self):
        token = Token(idx=0, text='w')
        counter = {'token_characters': {'w': 0}}
        self.single_id_indexer.count_vocab_items(token, counter)
        self.assertEqual(counter['token_characters']['w'], 1)

    def test_tokens_to_indices(self):
        self.assertEqual(
            self.single_id_indexer.tokens_to_indices(
                [Token('w'), Token('nawet')], self.vocabulary),
            {'tokens': [11, 127]})



"""
Custom character token indexer.
Adapted from COMBO.
Author: Mateusz Klimaszewski
"""
import itertools
from typing import List, Dict

import torch
from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.token_indexers import IndexedTokenList
from combo.data.token_indexers.token_characters_indexer import TokenCharactersIndexer
from combo.data.tokenizers import CharacterTokenizer
from combo.utils import pad_sequence_to_length


@Registry.register('characters_const_padding_token_indexer')
class TokenConstPaddingCharactersIndexer(TokenCharactersIndexer):
    """Wrapper around allennlp token indexer with const padding."""

    @register_arguments
    def __init__(self,
                 namespace: str = "token_characters",
                 tokenizer: CharacterTokenizer = CharacterTokenizer(),
                 start_tokens: List[str] = None,
                 end_tokens: List[str] = None,
                 min_padding_length: int = 0,
                 token_min_padding_length: int = 0):
        super().__init__(namespace, tokenizer, start_tokens, end_tokens, min_padding_length,
                         token_min_padding_length)

    @overrides
    def get_padding_lengths(self, indexed_tokens: IndexedTokenList) -> Dict[str, int]:
        padding_lengths = {"token_characters": len(indexed_tokens["token_characters"]),
                           "num_token_characters": self._min_padding_length}
        return padding_lengths

    @overrides
    def as_padded_tensor_dict(
            self, tokens: IndexedTokenList, padding_lengths: Dict[str, int]
    ) -> Dict[str, torch.Tensor]:
        # Pad the tokens.
        padded_tokens = pad_sequence_to_length(
            tokens["token_characters"],
            padding_lengths["token_characters"],
            default_value=lambda: [],
        )

        # Pad the characters within the tokens.
        desired_token_length = padding_lengths["num_token_characters"]
        longest_token: List[int] = max(tokens["token_characters"], key=len, default=[])  # type: ignore
        padding_value = 0
        if desired_token_length > len(longest_token):
            # Since we want to pad to greater than the longest token, we add a
            # "dummy token" so we can take advantage of the fast implementation of itertools.zip_longest.
            padded_tokens.append([padding_value] * desired_token_length)
        # pad the list of lists to the longest sublist, appending 0's
        padded_tokens = list(zip(*itertools.zip_longest(*padded_tokens, fillvalue=padding_value)))
        if desired_token_length > len(longest_token):
            # Removes the "dummy token".
            padded_tokens.pop()
        # Truncates all the tokens to the desired length, and return the result.
        return {
            "token_characters": torch.LongTensor(
                [list(token[:desired_token_length]) for token in padded_tokens]
            )
        }



from typing import Optional, List, Dict, Iterable

import torch
from overrides import overrides


"""
Class Metric adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/training/metrics/metric.py
"""


class Metric:
    """
    A very general abstract class representing a metric which can be
    accumulated.
    """

    supports_distributed = False

    def __call__(
        self, predictions: torch.Tensor, gold_labels: torch.Tensor, mask: Optional[torch.BoolTensor]
    ):
        """
        # Parameters
        predictions : `torch.Tensor`, required.
            A tensor of predictions.
        gold_labels : `torch.Tensor`, required.
            A tensor corresponding to some gold label to evaluate against.
        mask : `torch.BoolTensor`, optional (default = `None`).
            A mask can be passed, in order to deal with metrics which are
            computed over potentially padded elements, such as sequence labels.
        """
        raise NotImplementedError

    def get_metric(self, reset: bool):
        """
        Compute and return the metric. Optionally also call `self.reset`.
        """
        raise NotImplementedError

    def reset(self) -> None:
        """
        Reset any accumulators or internal state.
        """
        raise NotImplementedError

    @staticmethod
    def detach_tensors(*tensors: torch.Tensor) -> Iterable[torch.Tensor]:
        """
        If you actually passed gradient-tracking Tensors to a Metric, there will be
        a huge memory leak, because it will prevent garbage collection for the computation
        graph. This method ensures the tensors are detached.
        """
        # Check if it's actually a tensor in case something else was passed.
        return (x.detach() if isinstance(x, torch.Tensor) else x for x in tensors)


"""
Adapted from COMBO
Author: Mateusz Klimaszewski
"""


class LemmaAccuracy(Metric):

    def __init__(self):
        self._correct_count = 0.0
        self._total_count = 0.0
        self.correct_indices = torch.ones([])

    @overrides
    def __call__(self,
                 predictions: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: Optional[torch.BoolTensor] = None):
        if gold_labels is None:
            return
        predictions, gold_labels, mask = self.detach_tensors(predictions,
                                                             gold_labels,
                                                             mask)

        # Some sanity checks.
        if gold_labels.size() != predictions.size():
            raise ValueError(
                f"gold_labels must have shape == predictions.size() but "
                f"found tensor of shape: {gold_labels.size()}"
            )
        if mask is not None and mask.size() not in [predictions.size()[:-1], predictions.size()]:
            raise ValueError(
                f"mask must have shape in one of [predictions.size()[:-1], predictions.size()] but "
                f"found tensor of shape: {mask.size()}"
            )
        if mask is None:
            mask = predictions.new_ones(predictions.size()[:-1]).bool()
        if mask.dim() < predictions.dim():
            mask = mask.unsqueeze(-1)

        padding_mask = gold_labels.gt(0)
        correct = predictions.eq(gold_labels) * padding_mask
        correct = (correct.int().sum(-1) == padding_mask.int().sum(-1)) * mask.squeeze(-1)
        correct = correct.float()

        self.correct_indices = correct.flatten().bool()
        self._correct_count += correct.sum()
        self._total_count += mask.sum()

    @overrides
    def get_metric(self, reset: bool) -> float:
        if self._total_count > 0:
            accuracy = float(self._correct_count) / float(self._total_count)
        else:
            accuracy = 0.0
        if reset:
            self.reset()
        return accuracy

    @overrides
    def reset(self) -> None:
        self._correct_count = 0.0
        self._total_count = 0.0
        self.correct_indices = torch.ones([])


class SequenceBoolAccuracy(Metric):
    """BoolAccuracy implementation to handle sequences."""

    def __init__(self, prod_last_dim: bool = False):
        self._correct_count = 0.0
        self._total_count = 0.0
        self.prod_last_dim = prod_last_dim
        self.correct_indices = torch.ones([])

    @overrides
    def __call__(self,
                 predictions: torch.Tensor,
                 gold_labels: torch.Tensor,
                 mask: Optional[torch.BoolTensor] = None):
        if gold_labels is None:
            return
        predictions, gold_labels, mask = self.detach_tensors(predictions,
                                                             gold_labels,
                                                             mask)

        # Some sanity checks.
        if gold_labels.size() != predictions.size():
            raise ValueError(
                f"gold_labels must have shape == predictions.size() but "
                f"found tensor of shape: {gold_labels.size()}"
            )
        if mask is not None and mask.size() not in [predictions.size()[:-1], predictions.size()]:
            raise ValueError(
                f"mask must have shape in one of [predictions.size()[:-1], predictions.size()] but "
                f"found tensor of shape: {mask.size()}"
            )
        if mask is None:
            mask = predictions.new_ones(predictions.size()[:-1]).bool()
        if mask.dim() < predictions.dim():
            mask = mask.unsqueeze(-1)

        correct = predictions.eq(gold_labels) * mask

        if self.prod_last_dim:
            correct = correct.prod(-1).unsqueeze(-1)

        correct = correct.float()

        self.correct_indices = correct.flatten().bool()
        self._correct_count += correct.sum()
        self._total_count += mask.sum()

    @overrides
    def get_metric(self, reset: bool) -> float:
        if self._total_count > 0:
            accuracy = float(self._correct_count) / float(self._total_count)
        else:
            accuracy = 0.0
        if reset:
            self.reset()
        return accuracy

    @overrides
    def reset(self) -> None:
        self._correct_count = 0.0
        self._total_count = 0.0
        self.correct_indices = torch.ones([])


class AttachmentScores(Metric):
    """
    Computes labeled and unlabeled attachment scores for a
    dependency parse, as well as sentence level exact match
    for both labeled and unlabeled trees. Note that the input
    to this metric is the sampled predictions, not the distribution
    itself.

    # Parameters

    ignore_classes : `List[int]`, optional (default = None)
        A list of label ids to ignore when computing metrics.
    """

    def __init__(self, ignore_classes: List[int] = None) -> None:
        self._labeled_correct = 0.0
        self._unlabeled_correct = 0.0
        self._exact_labeled_correct = 0.0
        self._exact_unlabeled_correct = 0.0
        self._total_words = 0.0
        self._total_sentences = 0.0
        self.correct_indices = torch.ones([])

        self._ignore_classes: List[int] = ignore_classes or []

    def __call__(  # type: ignore
            self,
            predicted_indices: torch.Tensor,
            predicted_labels: torch.Tensor,
            gold_indices: torch.Tensor,
            gold_labels: torch.Tensor,
            mask: Optional[torch.BoolTensor] = None,
    ):
        """
        # Parameters

        predicted_indices : `torch.Tensor`, required.
            A tensor of head index predictions of shape (batch_size, timesteps).
        predicted_labels : `torch.Tensor`, required.
            A tensor of arc label predictions of shape (batch_size, timesteps).
        gold_indices : `torch.Tensor`, required.
            A tensor of the same shape as `predicted_indices`.
        gold_labels : `torch.Tensor`, required.
            A tensor of the same shape as `predicted_labels`.
        mask : `torch.BoolTensor`, optional (default = None).
            A tensor of the same shape as `predicted_indices`.
        """
        if gold_labels is None or gold_indices is None:
            return
        detached = self.detach_tensors(
            predicted_indices, predicted_labels, gold_indices, gold_labels, mask
        )
        predicted_indices, predicted_labels, gold_indices, gold_labels, mask = detached

        if mask is None:
            mask = torch.ones_like(predicted_indices).bool()

        predicted_indices = predicted_indices.long()
        predicted_labels = predicted_labels.long()
        gold_indices = gold_indices.long()
        gold_labels = gold_labels.long()

        # Multiply by a mask denoting locations of
        # gold labels which we should ignore.
        for label in self._ignore_classes:
            label_mask = gold_labels.eq(label)
            mask = mask & ~label_mask

        correct_indices = predicted_indices.eq(gold_indices).long() * mask
        unlabeled_exact_match = (correct_indices + ~mask).prod(dim=-1)
        if len(correct_indices.size()) > 2:
            unlabeled_exact_match = unlabeled_exact_match.prod(dim=-1)
        correct_labels = predicted_labels.eq(gold_labels).long() * mask
        correct_labels_and_indices = correct_indices * correct_labels
        self.correct_indices = correct_labels_and_indices.flatten()
        labeled_exact_match = (correct_labels_and_indices + ~mask).prod(dim=-1)
        if len(correct_indices.size()) > 2:
            labeled_exact_match = labeled_exact_match.prod(dim=-1)

        self._unlabeled_correct += correct_indices.sum()
        self._exact_unlabeled_correct += unlabeled_exact_match.sum()
        self._labeled_correct += correct_labels_and_indices.sum()
        self._exact_labeled_correct += labeled_exact_match.sum()
        self._total_sentences += correct_indices.size(0)
        self._total_words += correct_indices.numel() - (~mask).sum()

    def get_metric(self, reset: bool = False):
        """
        # Returns

        The accumulated metrics as a dictionary.
        """
        unlabeled_attachment_score = 0.0
        labeled_attachment_score = 0.0
        unlabeled_exact_match = 0.0
        labeled_exact_match = 0.0
        if self._total_words > 0.0:
            unlabeled_attachment_score = float(self._unlabeled_correct) / float(self._total_words)
            labeled_attachment_score = float(self._labeled_correct) / float(self._total_words)
        if self._total_sentences > 0:
            unlabeled_exact_match = float(self._exact_unlabeled_correct) / float(
                self._total_sentences
            )
            labeled_exact_match = float(self._exact_labeled_correct) / float(self._total_sentences)
        if reset:
            self.reset()
        return {
            "UAS": unlabeled_attachment_score,
            "LAS": labeled_attachment_score,
            "UEM": unlabeled_exact_match,
            "LEM": labeled_exact_match,
        }

    @overrides
    def reset(self):
        self._labeled_correct = 0.0
        self._unlabeled_correct = 0.0
        self._exact_labeled_correct = 0.0
        self._exact_unlabeled_correct = 0.0
        self._total_words = 0.0
        self._total_sentences = 0.0
        self.correct_indices = torch.ones([])


class SemanticMetrics(Metric):
    """Groups metrics for all predictions."""

    def __init__(self) -> None:
        self.upos_score = SequenceBoolAccuracy()
        self.xpos_score = SequenceBoolAccuracy()
        self.semrel_score = SequenceBoolAccuracy()
        self.feats_score = SequenceBoolAccuracy(prod_last_dim=True)
        self.lemma_score = LemmaAccuracy()
        self.attachment_scores = AttachmentScores()
        # Ignore PADDING and OOV
        self.enhanced_attachment_scores = AttachmentScores(ignore_classes=[0, 1])
        self.em_score = 0.0

    def __call__(  # type: ignore
            self,
            predictions: Dict[str, torch.Tensor],
            gold_labels: Dict[str, torch.Tensor],
            mask: torch.BoolTensor):
        self.upos_score(predictions["upostag"], gold_labels["upostag"], mask)
        self.xpos_score(predictions["xpostag"], gold_labels["xpostag"], mask)
        self.semrel_score(predictions["semrel"], gold_labels["semrel"], mask)
        self.feats_score(predictions["feats"], gold_labels["feats"], mask)
        self.lemma_score(predictions["lemma"], gold_labels["lemma"], mask)
        self.attachment_scores(predictions["head"],
                               predictions["deprel"],
                               gold_labels["head"],
                               gold_labels["deprel"],
                               mask)
        self.enhanced_attachment_scores(predictions["enhanced_head"],
                                        predictions["enhanced_deprel"],
                                        gold_labels["enhanced_head"],
                                        gold_labels["enhanced_deprel"],
                                        mask=None)
        enhanced_indices = (
            self.enhanced_attachment_scores.correct_indices.reshape(mask.size(0), mask.size(1) + 1, -1)[:, 1:, 1:].sum(
                -1).reshape(-1).bool()
            if len(self.enhanced_attachment_scores.correct_indices.size()) > 0
            else self.enhanced_attachment_scores.correct_indices
        )
        total = mask.sum()
        correct_indices = (self.upos_score.correct_indices *
                           self.xpos_score.correct_indices *
                           self.semrel_score.correct_indices *
                           self.feats_score.correct_indices *
                           self.lemma_score.correct_indices *
                           self.attachment_scores.correct_indices *
                           enhanced_indices) * mask.flatten()

        total, correct_indices = self.detach_tensors(total, correct_indices.float().sum())
        self.em_score = (correct_indices / total).item()

    def get_metric(self, reset: bool) -> Dict[str, float]:
        metrics_dict = {
            "UPOS_ACC": self.upos_score.get_metric(reset),
            "XPOS_ACC": self.xpos_score.get_metric(reset),
            "SEMREL_ACC": self.semrel_score.get_metric(reset),
            "LEMMA_ACC": self.lemma_score.get_metric(reset),
            "FEATS_ACC": self.feats_score.get_metric(reset),
            "EM": self.em_score
        }
        metrics_dict.update(self.attachment_scores.get_metric(reset))
        enhanced_metrics = {f"E{k}": v for k, v in self.enhanced_attachment_scores.get_metric(reset).items()}
        metrics_dict.update(enhanced_metrics)
        return metrics_dict

    def reset(self) -> None:
        self.upos_score.reset()
        self.xpos_score.reset()
        self.semrel_score.reset()
        self.lemma_score.reset()
        self.feats_score.reset()
        self.attachment_scores.reset()
        self.enhanced_attachment_scores.reset()
        self.em_score = 0.0




"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/modules/augmented_lstm.py
"""
from typing import Optional, Tuple

import torch
from torch.nn.utils.rnn import PackedSequence, pack_padded_sequence, pad_packed_sequence

from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments
from combo.modules.module import Module
from combo.nn.utils import get_dropout_mask

from combo.nn.initializers import block_orthogonal
from combo.utils import ConfigurationError


@Registry.register('augumented_lstm_cell')
class AugmentedLSTMCell(Module, FromParameters):
    """
    `AugmentedLSTMCell` implements a AugmentedLSTM cell.

    # Parameters

    embed_dim : `int`
        The number of expected features in the input.
    lstm_dim : `int`
        Number of features in the hidden state of the LSTM.
    use_highway : `bool`, optional (default = `True`)
        If `True` we append a highway network to the outputs of the LSTM.
    use_bias : `bool`, optional (default = `True`)
        If `True` we use a bias in our LSTM calculations, otherwise we don't.

    # Attributes

    input_linearity : `nn.Module`
        Fused weight matrix which computes a linear function over the input.
    state_linearity : `nn.Module`
        Fused weight matrix which computes a linear function over the states.
    """

    @register_arguments
    def __init__(
        self, embed_dim: int, lstm_dim: int, use_highway: bool = True, use_bias: bool = True
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.lstm_dim = lstm_dim
        self.use_highway = use_highway
        self.use_bias = use_bias

        if use_highway:
            self._highway_inp_proj_start = 5 * self.lstm_dim
            self._highway_inp_proj_end = 6 * self.lstm_dim

            # fused linearity of input to input_gate,
            # forget_gate, memory_init, output_gate, highway_gate,
            # and the actual highway value
            self.input_linearity = torch.nn.Linear(
                self.embed_dim, self._highway_inp_proj_end, bias=self.use_bias
            )
            # fused linearity of input to input_gate,
            # forget_gate, memory_init, output_gate, highway_gate
            self.state_linearity = torch.nn.Linear(
                self.lstm_dim, self._highway_inp_proj_start, bias=True
            )
        else:
            # If there's no highway layer then we have a standard
            # LSTM. The 4 comes from fusing input, forget, memory, output
            # gates/inputs.
            self.input_linearity = torch.nn.Linear(
                self.embed_dim, 4 * self.lstm_dim, bias=self.use_bias
            )
            self.state_linearity = torch.nn.Linear(self.lstm_dim, 4 * self.lstm_dim, bias=True)
        self.reset_parameters()

    def reset_parameters(self):
        # Use sensible default initializations for parameters.
        block_orthogonal(self.input_linearity.weight.data, [self.lstm_dim, self.embed_dim])
        block_orthogonal(self.state_linearity.weight.data, [self.lstm_dim, self.lstm_dim])

        self.state_linearity.bias.data.fill_(0.0)
        # Initialize forget gate biases to 1.0 as per An Empirical
        # Exploration of Recurrent Network Architectures, (Jozefowicz, 2015).
        self.state_linearity.bias.data[self.lstm_dim : 2 * self.lstm_dim].fill_(1.0)

    def forward(
        self,
        x: torch.Tensor,
        states=Tuple[torch.Tensor, torch.Tensor],
        variational_dropout_mask: Optional[torch.BoolTensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        !!! Warning
            DO NOT USE THIS LAYER DIRECTLY, instead use the AugmentedLSTM class

        # Parameters

        x : `torch.Tensor`
            Input tensor of shape (bsize x input_dim).
        states : `Tuple[torch.Tensor, torch.Tensor]`
            Tuple of tensors containing
            the hidden state and the cell state of each element in
            the batch. Each of these tensors have a dimension of
            (bsize x nhid). Defaults to `None`.

        # Returns

        `Tuple[torch.Tensor, torch.Tensor]`
            Returned states. Shape of each state is (bsize x nhid).

        """
        hidden_state, memory_state = states

        # In Pytext this was done as the last step of the cell.
        # But the original AugmentedLSTM from AllenNLP this was done before the processing
        if variational_dropout_mask is not None and self.training:
            hidden_state = hidden_state * variational_dropout_mask

        projected_input = self.input_linearity(x)
        projected_state = self.state_linearity(hidden_state)

        input_gate = forget_gate = memory_init = output_gate = highway_gate = None
        if self.use_highway:
            fused_op = projected_input[:, : 5 * self.lstm_dim] + projected_state
            fused_chunked = torch.chunk(fused_op, 5, 1)
            (input_gate, forget_gate, memory_init, output_gate, highway_gate) = fused_chunked
            highway_gate = torch.sigmoid(highway_gate)
        else:
            fused_op = projected_input + projected_state
            input_gate, forget_gate, memory_init, output_gate = torch.chunk(fused_op, 4, 1)
        input_gate = torch.sigmoid(input_gate)
        forget_gate = torch.sigmoid(forget_gate)
        memory_init = torch.tanh(memory_init)
        output_gate = torch.sigmoid(output_gate)
        memory = input_gate * memory_init + forget_gate * memory_state
        timestep_output: torch.Tensor = output_gate * torch.tanh(memory)

        if self.use_highway:
            highway_input_projection = projected_input[
                :, self._highway_inp_proj_start : self._highway_inp_proj_end
            ]
            timestep_output = (
                highway_gate * timestep_output
                + (1 - highway_gate) * highway_input_projection  # type: ignore
            )

        return timestep_output, memory


@Registry.register('augumented_lstm')
class AugmentedLstm(Module, FromParameters):
    """
    `AugmentedLstm` implements a one-layer single directional
    AugmentedLSTM layer. AugmentedLSTM is an LSTM which optionally
    appends an optional highway network to the output layer. Furthermore the
    dropout controls the level of variational dropout done.

    # Parameters

    input_size : `int`
        The number of expected features in the input.
    hidden_size : `int`
        Number of features in the hidden state of the LSTM.
        Defaults to 32.
    go_forward : `bool`
        Whether to compute features left to right (forward)
        or right to left (backward).
    recurrent_dropout_probability : `float`
        Variational dropout probability to use. Defaults to 0.0.
    use_highway : `bool`
        If `True` we append a highway network to the outputs of the LSTM.
    use_input_projection_bias : `bool`
        If `True` we use a bias in our LSTM calculations, otherwise we don't.

    # Attributes

    cell : `AugmentedLSTMCell`
        `AugmentedLSTMCell` that is applied at every timestep.

    """

    @register_arguments
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        go_forward: bool = True,
        recurrent_dropout_probability: float = 0.0,
        use_highway: bool = True,
        use_input_projection_bias: bool = True,
    ):
        super().__init__()

        self.embed_dim = input_size
        self.lstm_dim = hidden_size

        self.go_forward = go_forward
        self.use_highway = use_highway
        self.recurrent_dropout_probability = recurrent_dropout_probability

        self.cell = AugmentedLSTMCell(
            self.embed_dim, self.lstm_dim, self.use_highway, use_input_projection_bias
        )

    def forward(
        self, inputs: PackedSequence, states: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
    ) -> Tuple[PackedSequence, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Warning: Would be better to use the BiAugmentedLstm class in a regular model

        Given an input batch of sequential data such as word embeddings, produces a single layer unidirectional
        AugmentedLSTM representation of the sequential input and new state tensors.

        # Parameters

        inputs : `PackedSequence`
            `bsize` sequences of shape `(len, input_dim)` each, in PackedSequence format
        states : `Tuple[torch.Tensor, torch.Tensor]`
            Tuple of tensors containing the initial hidden state and
            the cell state of each element in the batch. Each of these tensors have a dimension of
            (1 x bsize x nhid). Defaults to `None`.

        # Returns

        `Tuple[PackedSequence, Tuple[torch.Tensor, torch.Tensor]]`
            AugmentedLSTM representation of input and the state of the LSTM `t = seq_len`.
            Shape of representation is (bsize x seq_len x representation_dim).
            Shape of each state is (1 x bsize x nhid).

        """
        if not isinstance(inputs, PackedSequence):
            raise ConfigurationError("inputs must be PackedSequence but got %s" % (type(inputs)))

        sequence_tensor, batch_lengths = pad_packed_sequence(inputs, batch_first=True)
        batch_size = sequence_tensor.size()[0]
        total_timesteps = sequence_tensor.size()[1]
        output_accumulator = sequence_tensor.new_zeros(batch_size, total_timesteps, self.lstm_dim)
        if states is None:
            full_batch_previous_memory = sequence_tensor.new_zeros(batch_size, self.lstm_dim)
            full_batch_previous_state = sequence_tensor.data.new_zeros(batch_size, self.lstm_dim)
        else:
            full_batch_previous_state = states[0].squeeze(0)
            full_batch_previous_memory = states[1].squeeze(0)
        current_length_index = batch_size - 1 if self.go_forward else 0
        if self.recurrent_dropout_probability > 0.0:
            dropout_mask = get_dropout_mask(
                self.recurrent_dropout_probability, full_batch_previous_memory
            )
        else:
            dropout_mask = None

        for timestep in range(total_timesteps):
            index = timestep if self.go_forward else total_timesteps - timestep - 1

            if self.go_forward:
                while batch_lengths[current_length_index] <= index:
                    current_length_index -= 1
            # If we're going backwards, we are _picking up_ more indices.
            else:
                # First conditional: Are we already at the maximum
                # number of elements in the batch?
                # Second conditional: Does the next shortest
                # sequence beyond the current batch
                # index require computation use this timestep?
                while (
                    current_length_index < (len(batch_lengths) - 1)
                    and batch_lengths[current_length_index + 1] > index
                ):
                    current_length_index += 1

            previous_memory = full_batch_previous_memory[0 : current_length_index + 1].clone()
            previous_state = full_batch_previous_state[0 : current_length_index + 1].clone()
            timestep_input = sequence_tensor[0 : current_length_index + 1, index]
            timestep_output, memory = self.cell(
                timestep_input,
                (previous_state, previous_memory),
                dropout_mask[0 : current_length_index + 1] if dropout_mask is not None else None,
            )
            full_batch_previous_memory = full_batch_previous_memory.data.clone()
            full_batch_previous_state = full_batch_previous_state.data.clone()
            full_batch_previous_memory[0 : current_length_index + 1] = memory
            full_batch_previous_state[0 : current_length_index + 1] = timestep_output
            output_accumulator[0 : current_length_index + 1, index, :] = timestep_output

        output_accumulator = pack_padded_sequence(
            output_accumulator, batch_lengths, batch_first=True
        )

        # Mimic the pytorch API by returning state in the following shape:
        # (num_layers * num_directions, batch_size, lstm_dim). As this
        # LSTM cannot be stacked, the first dimension here is just 1.
        final_state = (
            full_batch_previous_state.unsqueeze(0),
            full_batch_previous_memory.unsqueeze(0),
        )
        return output_accumulator, final_state



import json
import logging
import os
import pathlib
import tempfile
from itertools import chain
from typing import Dict, Optional, Any, Tuple

import torch
from absl import app, flags
import pytorch_lightning as pl
from pytorch_lightning.loggers import TensorBoardLogger
from tqdm import tqdm

from combo.training.trainable_combo import TrainableCombo
from combo.utils import checks, ComboLogger

from combo.config import resolve
from combo.default_model import default_ud_dataset_reader, default_data_loader, default_model
from combo.modules.archival import load_archive, archive
from combo.predict import COMBO
from combo.data import api
from combo.config import override_parameters
from combo.data import LamboTokenizer, Vocabulary, DatasetReader
from combo.data.dataset_loaders import DataLoader
from combo.modules.model import Model
from combo.utils import ConfigurationError
from combo.utils.matrices import extract_combo_matrices

logging.setLoggerClass(ComboLogger)
logger = logging.getLogger(__name__)
_FEATURES = ["token", "char", "upostag", "xpostag", "lemma", "feats"]
_TARGETS = ["deprel", "feats", "head", "lemma", "upostag", "xpostag", "semrel", "sent", "deps"]


def handle_error(error: Exception, prefix: str):
    msg = getattr(error, 'message', str(error))
    logger.error(msg, prefix)


FLAGS = flags.FLAGS
flags.DEFINE_enum(name="mode", default=None, enum_values=["train", "predict"],
                  help="Specify COMBO mode: train or predict")

# Common flags
flags.DEFINE_integer(name="n_cuda_devices", default=-1,
                     help="Number of devices to train on (default -1 auto mode - train on as many as possible)")
flags.DEFINE_string(name="output_file", default="output.log",
                    help="Predictions result file.")

# Training flags
flags.DEFINE_string(name="training_data_path", default="", help="Training data path(s)")
flags.DEFINE_alias(name="training_data", original_name="training_data_path")
flags.DEFINE_string(name="validation_data_path", default="", help="Validation data path(s)")
flags.DEFINE_alias(name="validation_data", original_name="validation_data_path")
flags.DEFINE_string(name="vocabulary_path", default=None,
                    help="Stored vocabulary files. If not provided in training mode, vocabulary is built from training files")
flags.DEFINE_integer(name="lemmatizer_embedding_dim", default=300,
                     help="Lemmatizer embeddings dim")
flags.DEFINE_integer(name="num_epochs", default=400,
                     help="Epochs num")
flags.DEFINE_integer(name="batch_size", default=256,
                     help="Batch size")
flags.DEFINE_integer(name="batches_per_epoch", default=16,
                     help="Number of batches per epoch")
flags.DEFINE_string(name="pretrained_transformer_name", default="bert-base-cased",
                    help="Pretrained transformer model name (see transformers from HuggingFace library for list of "
                         "available models) for transformers based embeddings.")
flags.DEFINE_list(name="features", default=["token", "char"],
                  help=f"Features used to train model (required 'token' and 'char'). Possible values: {_FEATURES}.")
flags.DEFINE_list(name="targets", default=["deprel", "feats", "head", "lemma", "upostag", "xpostag"],
                  help=f"Targets of the model (required `deprel` and `head`). Possible values: {_TARGETS}.")
flags.DEFINE_string(name="serialization_dir", default=None,
                    help="Model serialization directory (default - system temp dir).")
flags.DEFINE_boolean(name="tensorboard", default=False,
                     help="When provided model will log tensorboard metrics.")
flags.DEFINE_string(name="tensorboard_name", default="combo",
                    help="Name of the model in TensorBoard logs.")
flags.DEFINE_string(name="config_path", default="",
                    help="Config file path.")
flags.DEFINE_list(name="datasets_for_vocabulary", default=["train"],
                  help="")
flags.DEFINE_boolean(name="turns", default=False,
                     help="Segment into sentences on sentence break or on turn break.")
flags.DEFINE_boolean(name="split_subwords", default=False,
                     help="Split subwords (e.g. don\'t = do, n\'t) into separate tokens.")

# Finetune after training flags
flags.DEFINE_string(name="finetuning_training_data_path", default="",
                    help="Training data path(s)")
flags.DEFINE_string(name="finetuning_validation_data_path", default="",
                    help="Validation data path(s)")

# Test after training flags
flags.DEFINE_string(name="test_data_path", default=None,
                    help="Test path file.")
flags.DEFINE_alias(name="test_data", original_name="test_data_path")
flags.DEFINE_boolean(name="use_pure_config", default=False,
                     help="Ignore ext flags.")

# Prediction flags
flags.DEFINE_string(name="model_path", default=None,
                    help="Pretrained model path.")
flags.DEFINE_string(name="input_file", default=None,
                    help="File to predict path")
flags.DEFINE_boolean(name="conllu_format", default=True,
                     help="Prediction based on conllu format (instead of raw text).")
flags.DEFINE_boolean(name="finetuning", default=False,
                     help="Finetuning mode for training.")
flags.DEFINE_string(name="tokenizer_language", default="English",
                    help="Tokenizer language.")
flags.DEFINE_boolean(name="save_matrices", default=True,
                     help="Save relation distribution matrices.")
flags.DEFINE_enum(name="predictor_name", default="combo-lambo",
                  enum_values=["combo", "combo-spacy", "combo-lambo"],
                  help="Use predictor with whitespace, spacy or lambo (recommended) tokenizer.")


def build_vocabulary_from_instances(training_data_loader: DataLoader,
                                    validation_data_loader: Optional[DataLoader],
                                    logging_prefix: str) -> Vocabulary:
    logger.info('Building a vocabulary from instances.', prefix=logging_prefix)
    if "train" in FLAGS.datasets_for_vocabulary and "valid" in FLAGS.datasets_for_vocabulary:
        instances = chain(training_data_loader.iter_instances(),
                          validation_data_loader.iter_instances()) \
            if validation_data_loader \
            else training_data_loader.iter_instances()
    elif "train" in FLAGS.datasets_for_vocabulary:
        instances = training_data_loader.iter_instances()
    elif "valid" in FLAGS.datasets_for_vocabulary:
        instances = validation_data_loader.iter_instances()
    else:
        logger.error("train and valid are the only allowed values for --datasets_for_vocabulary!",
                     prefix=logging_prefix)
        raise ValueError("train and valid are the only allowed values for --datasets_for_vocabulary!")
    vocabulary = Vocabulary.from_instances_extended(
        instances,
        non_padded_namespaces=['head_labels'],
        only_include_pretrained_words=False,
        oov_token='_',
        padding_token='__PAD__'
    )
    return vocabulary


def get_defaults(dataset_reader: Optional[DatasetReader],
                 training_data_loader: Optional[DataLoader],
                 validation_data_loader: Optional[DataLoader],
                 vocabulary: Optional[Vocabulary],
                 training_data_path: str,
                 validation_data_path: str,
                 prefix: str) -> Tuple[DatasetReader, DataLoader, DataLoader, Vocabulary]:
    if not dataset_reader and (FLAGS.test_data_path
                               or not training_data_loader
                               or (FLAGS.validation_data_path and not validation_data_loader)):
        # Dataset reader is required to read training data and/or for training (and validation) data loader
        dataset_reader = default_ud_dataset_reader(FLAGS.pretrained_transformer_name,
                                                   tokenizer=LamboTokenizer(FLAGS.tokenizer_language,
                                                                            default_turns=FLAGS.turns,
                                                                            default_split_subwords=FLAGS.split_subwords)
                                                   )

    if not training_data_loader:
        training_data_loader = default_data_loader(dataset_reader, training_data_path)
    else:
        if training_data_path:
            training_data_loader.data_path = training_data_path
        else:
            logger.warning(f'No training data path provided - using the path from configuration: ' +
                           str(training_data_loader.data_path), prefix=prefix)

    if FLAGS.validation_data_path and not validation_data_loader:
        validation_data_loader = default_data_loader(dataset_reader, validation_data_path)
    elif FLAGS.validation_data_path and validation_data_loader:
        if validation_data_path:
            validation_data_loader.data_path = validation_data_path
        else:
            logger.warning(f'No validation data path provided - using the path from configuration: ' +
                           str(validation_data_loader.data_path), prefix=prefix)

    if not vocabulary:
        vocabulary = build_vocabulary_from_instances(training_data_loader, validation_data_loader, prefix)

    return dataset_reader, training_data_loader, validation_data_loader, vocabulary


def _read_property_from_config(property_key: str,
                               params: Dict[str, Any],
                               logging_prefix: str,
                               pass_down_parameters: Dict[str, Any] = None) -> Optional[Any]:
    property = None
    pass_down_parameters = pass_down_parameters or {}
    if property_key in params:
        logger.info(f'Reading {property_key.replace("_", " ")} from parameters.', prefix=logging_prefix)
        try:
            property = resolve(params[property_key], pass_down_parameters=pass_down_parameters)
        except Exception as e:
            handle_error(e, logging_prefix)
    return property


def read_dataset_reader_from_config(params: Dict[str, Any],
                                    logging_prefix: str,
                                    pass_down_parameters: Dict[str, Any] = None) -> Optional[DataLoader]:
    return _read_property_from_config('dataset_reader', params, logging_prefix, pass_down_parameters)


def read_data_loader_from_config(params: Dict[str, Any],
                                 logging_prefix: str,
                                 validation: bool = False,
                                 pass_down_parameters: Dict[str, Any] = None) -> Optional[DataLoader]:
    key = 'validation_data_loader' if validation else 'data_loader'
    return _read_property_from_config(key, params, logging_prefix, pass_down_parameters)


def read_vocabulary_from_config(params: Dict[str, Any],
                                logging_prefix: str,
                                pass_down_parameters: Dict[str, Any] = None) -> Optional[Vocabulary]:
    vocabulary = None
    pass_down_parameters = pass_down_parameters or {}
    if "vocabulary" in params:
        logger.info('Reading vocabulary from saved directory.', prefix=logging_prefix)
        if 'directory' in params['vocabulary']['parameters']:
            params['vocabulary']['parameters']['directory'] = os.path.join('/'.join(FLAGS.config_path.split('/')[:-1]),
                                                                           params['vocabulary']['parameters'][
                                                                               'directory'])
        try:
            vocabulary = resolve(params['vocabulary'], pass_down_parameters)
        except Exception as e:
            handle_error(e, logging_prefix)
    return vocabulary


def read_model_from_config(logging_prefix: str) -> Optional[
    Tuple[Model, DatasetReader, DataLoader, DataLoader, Vocabulary]]:
    try:
        checks.file_exists(FLAGS.config_path)
    except ConfigurationError as e:
        handle_error(e, logging_prefix)
        return

    if FLAGS.serialization_dir is None:
        logger.error(f'--serialization_dir was not passed as an argument!')
        print(f'--serialization_dir was not passed as an argument!')
        return

    with open(FLAGS.config_path, 'r') as f:
        params = json.load(f)

    if not FLAGS.use_pure_config:
        params = override_parameters(params, _get_ext_vars(finetuning=False))
        if 'feats' not in FLAGS.targets and 'morphological_feat' in params['model']['parameters']:
            del params['model']['parameters']['morphological_feat']

    pass_down_parameters = {}
    model_name = params.get("model_name")
    if model_name:
        pass_down_parameters = {"model_name": model_name}

    dataset_reader = read_dataset_reader_from_config(params, logging_prefix, pass_down_parameters)
    training_data_loader = read_data_loader_from_config(params, logging_prefix,
                                                        validation=False, pass_down_parameters=pass_down_parameters)
    if (
            not FLAGS.validation_data_path or not FLAGS.finetuning_validation_data_path) and 'validation_data_loader' in params:
        logger.warning('Validation data loader is in parameters, but no validation data path was provided!')
        validation_data_loader = None
    else:
        validation_data_loader = read_data_loader_from_config(params, logging_prefix,
                                                              validation=True,
                                                              pass_down_parameters=pass_down_parameters)
    vocabulary = read_vocabulary_from_config(params, logging_prefix, pass_down_parameters)

    dataset_reader, training_data_loader, validation_data_loader, vocabulary = get_defaults(
        dataset_reader,
        training_data_loader,
        validation_data_loader,
        vocabulary,
        FLAGS.training_data_path if not FLAGS.finetuning else FLAGS.finetuning_training_data_path,
        FLAGS.validation_data_path if not FLAGS.finetuning else FLAGS.finetuning_validation_data_path,
        logging_prefix
    )

    pass_down_parameters = {'vocabulary': vocabulary}
    if not FLAGS.use_pure_config:
        pass_down_parameters['model_name'] = FLAGS.pretrained_transformer_name

    logger.info('Resolving the model from parameters.', prefix=logging_prefix)
    model = resolve(params['model'], pass_down_parameters=pass_down_parameters)

    return model, dataset_reader, training_data_loader, validation_data_loader, vocabulary


def run(_):
    if FLAGS.mode == 'train':
        model, dataset_reader, training_data_loader, validation_data_loader, vocabulary = None, None, None, None, None

        if not FLAGS.finetuning:
            prefix = 'Training'
            logger.info('Setting up the model for training', prefix=prefix)

            if FLAGS.config_path:
                logger.info(f'Reading parameters from configuration path {FLAGS.config_path}', prefix=prefix)
                model, dataset_reader, training_data_loader, validation_data_loader, vocabulary = read_model_from_config(
                    prefix)
            else:
                dataset_reader, training_data_loader, validation_data_loader, vocabulary = get_defaults(
                    dataset_reader,
                    training_data_loader,
                    validation_data_loader,
                    vocabulary,
                    FLAGS.training_data_path,
                    FLAGS.validation_data_path,
                    prefix
                )
                model = default_model(FLAGS.pretrained_transformer_name, vocabulary)

            if FLAGS.use_pure_config and model is None:
                logger.error('Error in configuration - model could not be read from parameters. ' +
                             'Correct the configuration or use --nopure_config ' +
                             'to use default models.')
                return

            try:
                pathlib.Path(FLAGS.serialization_dir).mkdir(parents=True, exist_ok=True)
            except FileExistsError:
                pass
            serialization_dir = tempfile.mkdtemp(prefix='combo', dir=FLAGS.serialization_dir)

        else:
            prefix = 'Finetuning'

            try:
                checks.file_exists(FLAGS.finetuning_training_data_path)
                if FLAGS.finetuning_validation_data_path:
                    checks.file_exists(FLAGS.finetuning_validation_data_path)
            except ConfigurationError as e:
                handle_error(e, prefix)

            logger.info('Loading the model for finetuning', prefix=prefix)
            model, _, training_data_loader, validation_data_loader, dataset_reader = load_archive(FLAGS.model_path)

            if model is None:
                logger.error(f'Model could not be loaded from archive {str(FLAGS.model_path)}. Exiting', prefix=prefix)
                return

            vocabulary = model.vocab

            try:
                pathlib.Path(FLAGS.serialization_dir).mkdir(parents=True, exist_ok=True)
            except FileExistsError:
                pass
            serialization_dir = tempfile.mkdtemp(prefix='combo', suffix='-finetuning', dir=FLAGS.serialization_dir)

            dataset_reader, training_data_loader, validation_data_loader, vocabulary = get_defaults(
                dataset_reader,
                training_data_loader,
                validation_data_loader,
                vocabulary,
                FLAGS.finetuning_training_data_path,
                FLAGS.finetuning_validation_data_path,
                prefix
            )

        logger.info('Indexing training data loader', prefix=prefix)
        training_data_loader.index_with(model.vocab)

        if validation_data_loader:
            logger.info('Indexing validation data loader', prefix=prefix)
            validation_data_loader.index_with(model.vocab)

        nlp = TrainableCombo(model,
                             torch.optim.Adam,
                             optimizer_kwargs={'betas': [0.9, 0.9], 'lr': 0.002},
                             validation_metrics=['EM'])

        n_cuda_devices = "auto" if FLAGS.n_cuda_devices == -1 else FLAGS.n_cuda_devices

        tensorboard_logger = TensorBoardLogger(os.path.join(serialization_dir, 'tensorboard_logs'),
                                               name=FLAGS.tensorboard_name) if FLAGS.tensorboard else None

        trainer = pl.Trainer(max_epochs=FLAGS.num_epochs,
                             default_root_dir=serialization_dir,
                             gradient_clip_val=5,
                             devices=n_cuda_devices,
                             logger=tensorboard_logger)
        trainer.fit(model=nlp, train_dataloaders=training_data_loader, val_dataloaders=validation_data_loader)

        logger.info(f'Archiving the model in {serialization_dir}', prefix=prefix)
        archive(model, serialization_dir, training_data_loader, validation_data_loader, dataset_reader)
        logger.info(f"Model stored in: {serialization_dir}", prefix=prefix)

    elif FLAGS.mode == 'predict':
        prefix = 'Predicting'
        logger.info('Loading the model', prefix=prefix)
        model, config, _, _, dataset_reader = load_archive(FLAGS.model_path)

        if config.get("tokenizer_language") is None:
            logger.warning("Tokenizer language was not found in archive's configuration file - " +
                           "using the --tokenizer_language parameter (default: English)")
        tokenizer_language = config.get("tokenizer_language", FLAGS.tokenizer_language)

        if not dataset_reader:
            logger.info("No dataset reader in the configuration or archive file - using a default UD dataset reader",
                        prefix=prefix)
            dataset_reader = default_ud_dataset_reader(FLAGS.pretrained_transformer_name,
                                                       tokenizer=LamboTokenizer(tokenizer_language,
                                                                                default_turns=FLAGS.turns,
                                                                                default_split_subwords=FLAGS.split_subwords)
                                                       )

        predictor = COMBO(model, dataset_reader)

        if FLAGS.input_file == '-':
            print("Interactive mode.")
            sentence = input("Sentence: ")
            prediction = [p.tokens for p in predictor(sentence)]
            # Flatten the prediction
            flattened_prediction = []
            for p in prediction:
                flattened_prediction.extend(p)
            print("{:15} {:15} {:10} {:10} {:10}".format('TOKEN', 'LEMMA', 'UPOS', 'HEAD', 'DEPREL'))
            for token in flattened_prediction:
                print("{:15} {:15} {:10} {:10} {:10}".format(token.text, token.lemma, token.upostag, token.head,
                                                             token.deprel))
        elif FLAGS.output_file:
            try:
                checks.file_exists(FLAGS.input_file)
            except ConfigurationError as e:
                handle_error(e, prefix)

            pathlib.Path(FLAGS.output_file).touch(exist_ok=True)

            logger.info("Predicting examples from file", prefix=prefix)

            predictions = []
            if FLAGS.conllu_format:
                test_trees = dataset_reader.read(FLAGS.input_file)
                with open(FLAGS.output_file, "w") as file:
                    for tree in tqdm(test_trees):
                        prediction = predictor.predict_instance(tree)
                        file.writelines(api.serialize_token_list(api.sentence2conllu(prediction,
                                                                                     keep_semrel=dataset_reader.use_sem)))
                        predictions.append(prediction)

            else:
                tokenizer = LamboTokenizer(tokenizer_language)
                with open(FLAGS.input_file, "r", encoding='utf-8') as file:
                    input_sentences = tokenizer.segment(file.read(),
                                                        turns=FLAGS.turns,
                                                        split_subwords=FLAGS.split_subwords)
                predictions = predictor.predict(input_sentences)
                with open(FLAGS.output_file, "w") as file:
                    for prediction in tqdm(predictions):
                        file.writelines(api.serialize_token_list(api.sentence2conllu(prediction,
                                                                                     keep_semrel=dataset_reader.use_sem)))

            if FLAGS.save_matrices:
                logger.info("Saving matrices", prefix=prefix)
                if FLAGS.serialization_dir is None or pathlib.Path(FLAGS.serialization_dir).exists():
                    logger.warning('Serialization path was not passed as an argument - skipping matrix extraction.')
                    return
                extract_combo_matrices(predictions,
                                       pathlib.Path(FLAGS.serialization_dir),
                                       pathlib.Path(FLAGS.input_file),
                                       logger,
                                       prefix)

        else:
            msg = 'No output file for input file {input_file} specified.'.format(input_file=FLAGS.input_file)
            logger.info(msg, prefix=prefix)
            print(msg)


def _get_ext_vars(finetuning: bool = False) -> Dict:
    if FLAGS.use_pure_config:
        return {}

    to_override = {
        "model_name": FLAGS.pretrained_transformer_name,
        "model": {
            "parameters": {
                "lemmatizer": {
                    "parameters": {
                        "embedding_dim": FLAGS.lemmatizer_embedding_dim
                    }
                },
                "serialization_dir": FLAGS.serialization_dir
            }
        },
        "data_loader": {
            "parameters": {
                "data_path": FLAGS.training_data_path if not finetuning else FLAGS.finetuning_training_data_path,
                "batch_size": FLAGS.batch_size,
                "batches_per_epoch": FLAGS.batches_per_epoch,
                "reader": {
                    "parameters": {
                        "features": FLAGS.features,
                        "targets": FLAGS.targets,
                        "tokenizer": {
                            "parameters": {
                                "language": FLAGS.tokenizer_language
                            }
                        }
                    }
                }
            }
        },
        "validation_data_loader": {
            "parameters": {
                "data_path": FLAGS.validation_data_path if not finetuning else FLAGS.finetuning_validation_data_path,
                "batch_size": FLAGS.batch_size,
                "batches_per_epoch": FLAGS.batches_per_epoch,
                "reader": {
                    "parameters": {
                        "features": FLAGS.features,
                        "targets": FLAGS.targets,
                        "tokenizer": {
                            "parameters": {
                                "language": FLAGS.tokenizer_language
                            }
                        }
                    }
                }
            }
        },
        "dataset_reader": {
            "parameters": {
                "features": FLAGS.features,
                "targets": FLAGS.targets,
                "tokenizer": {
                    "parameters": {
                        "language": FLAGS.tokenizer_language
                    }
                }
            }
        }
    }

    if FLAGS.vocabulary_path:
        to_override["vocabulary"] = {
            "type": "from_files_vocabulary",
            "parameters": {
                "directory": FLAGS.vocabulary_path,
                "padding_token": "__PAD__",
                "oov_token": "_"
            }
        }
    return to_override


def main():
    """Parse flags."""
    flags.register_validator(
        "features",
        lambda values: all(
            value in _FEATURES for value in values),
        message="Flags --features contains unknown value(s)."
    )
    flags.register_validator(
        "mode",
        lambda value: value is not None,
        message="Flag --mode must be set with either `predict` or `train` value")
    flags.register_validator(
        "targets",
        lambda values: all(
            value in _TARGETS for value in values),
        message="Flag --targets contains unknown value(s)."
    )
    app.run(run)


if __name__ == "__main__":
    main()



"""
Adapted from COMBO
Author: Mateusz Klimaszewski
"""
from typing import Dict, List, Optional, Union
import torch
from overrides import overrides

from combo import data
from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.vocabulary import get_slices_if_not_provided
from combo.nn import base
from combo.nn.activations import Activation
from combo.nn.utils import masked_cross_entropy
from combo.predictors.predictor import Predictor
from combo.utils import ConfigurationError


@Registry.register('combo_morpho_from_vocab')
class MorphologicalFeatures(Predictor):
    """Morphological features predicting model."""

    @register_arguments
    def __init__(self,
                 vocabulary: data.Vocabulary,
                 vocab_namespace: str,
                 input_dim: int,
                 num_layers: int,
                 hidden_dims: List[int],
                 activations: List[Activation],  # change to Union[Activation, List[Activation]]
                 dropout: Union[float, List[float]] = 0.0,
                 ):
        super().__init__()
        if len(hidden_dims) + 1 != num_layers:
            raise ConfigurationError(
                f"len(hidden_dims) ({len(hidden_dims):d}) + 1 != num_layers ({num_layers:d})"
            )

        assert vocab_namespace in vocabulary.get_namespaces()

        slices = get_slices_if_not_provided(vocabulary)
        hidden_dims = hidden_dims + [vocabulary.get_vocab_size(vocab_namespace)]
        self.feedforward_network = base.FeedForward(
            input_dim=input_dim,
            num_layers=num_layers,
            hidden_dims=hidden_dims,
            activations=activations,
            dropout=dropout)
        self.slices = slices

    @overrides
    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        if mask is None:
            mask = x.new_ones(x.size()[:-1])

        x, feature_maps = self.feedforward_network(x)

        prediction = []
        for _, cat_indices in self.slices.items():
            prediction.append(x[:, :, cat_indices].argmax(dim=-1))

        output = {
            "prediction": torch.stack(prediction, dim=-1),
            "probability": x,
            "embedding": feature_maps[-1],
        }

        if labels is not None:
            if sample_weights is None:
                sample_weights = labels.new_ones([mask.size(0)])
            output["loss"] = self._loss(x, labels, mask, sample_weights)

        return output

    def _loss(self, pred: torch.Tensor, true: torch.Tensor, mask: torch.BoolTensor,
              sample_weights: torch.Tensor) -> torch.Tensor:
        assert pred.size() == true.size()
        BATCH_SIZE, _, MORPHOLOGICAL_FEATURES = pred.size()

        valid_positions = mask.sum()

        pred = pred.reshape(-1, MORPHOLOGICAL_FEATURES)
        true = true.reshape(-1, MORPHOLOGICAL_FEATURES)
        mask = mask.reshape(-1)
        loss = None
        loss_func = masked_cross_entropy
        for cat, cat_indices in self.slices.items():
            if cat not in ["__PAD__", "_"]:
                if loss is None:
                    loss = loss_func(pred[:, cat_indices],
                                     true[:, cat_indices].argmax(dim=1),
                                     mask)
                else:
                    loss += loss_func(pred[:, cat_indices],
                                      true[:, cat_indices].argmax(dim=1),
                                      mask)
        loss = loss.reshape(BATCH_SIZE, -1) * sample_weights.unsqueeze(-1)
        return loss.sum() / valid_positions



class InvalidTagSequence(Exception):
    def __init__(self, tag_sequence=None):
        super().__init__()
        self.tag_sequence = tag_sequence

    def __str__(self):
        return " ".join(self.tag_sequence)



"""
Adapted from AllenNLP.
https://github.com/allenai/allennlp/blob/main/allennlp/data/token_indexers/token_characters_indexer.py
"""

from typing import Dict, List
import itertools
import warnings

from overrides import overrides
import torch

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data import Vocabulary
from combo.data.token_indexers import TokenIndexer, IndexedTokenList
from combo.data.tokenizers import Token, CharacterTokenizer
from combo.utils import ConfigurationError, pad_sequence_to_length


@Registry.register('token_characters_indexer')
class TokenCharactersIndexer(TokenIndexer):
    """
    This :class:`TokenIndexer` represents tokens as lists of character indices.

    Registered as a `TokenIndexer` with name "characters".

    # Parameters

    _namespace : `str`, optional (default=`token_characters`)
        We will use this _namespace in the :class:`Vocabulary` to map the characters in each token
        to indices.
    tokenizer : `CharacterTokenizer`, optional (default=`CharacterTokenizer()`)
        We use a :class:`CharacterTokenizer` to handle splitting tokens into characters, as it has
        options for byte encoding and other things.  The default here is to instantiate a
        `CharacterTokenizer` with its default parameters, which uses unicode characters and
        retains casing.
    start_tokens : `List[str]`, optional (default=`None`)
        These are prepended to the tokens provided to `tokens_to_indices`.
    end_tokens : `List[str]`, optional (default=`None`)
        These are appended to the tokens provided to `tokens_to_indices`.
    min_padding_length : `int`, optional (default=`0`)
        We use this value as the minimum length of padding. Usually used with :class:`CnnEncoder`, its
        value should be set to the maximum value of `ngram_filter_sizes` correspondingly.
    token_min_padding_length : `int`, optional (default=`0`)
        See :class:`TokenIndexer`.
    """

    @register_arguments
    def __init__(
        self,
        namespace: str = "token_characters",
        tokenizer: CharacterTokenizer = CharacterTokenizer(),
        start_tokens: List[str] = None,
        end_tokens: List[str] = None,
        min_padding_length: int = 0,
        token_min_padding_length: int = 0,
    ) -> None:
        super().__init__(token_min_padding_length)
        if min_padding_length == 0:
            url = "https://github.com/allenai/allennlp/issues/1954"
            warnings.warn(
                "You are using the default value (0) of `min_padding_length`, "
                f"which can cause some subtle bugs (more info see {url}). "
                "Strongly recommend to set a value, usually the maximum size "
                "of the convolutional layer size when using CnnEncoder.",
                UserWarning,
            )
        self._min_padding_length = min_padding_length
        self._namespace = namespace
        self._character_tokenizer = tokenizer

        self._start_tokens = [Token(st) for st in (start_tokens or [])]
        self._end_tokens = [Token(et) for et in (end_tokens or [])]

    @overrides
    def count_vocab_items(self, token: Token, counter: Dict[str, Dict[str, int]]):
        if token.text is None:
            raise ConfigurationError("TokenCharactersIndexer needs a tokenizer that retains text")
        for character in self._character_tokenizer.tokenize(token.text):
            # If `text_id` is set on the character token (e.g., if we're using byte encoding), we
            # will not be using the vocabulary for this character.
            if getattr(character, "text_id", None) is None:
                counter[self._namespace][character.text] += 1

    @overrides
    def tokens_to_indices(
        self, tokens: List[Token], vocabulary: Vocabulary
    ) -> IndexedTokenList:
        indices: List[List[int]] = []
        for token in itertools.chain(self._start_tokens, tokens, self._end_tokens):
            token_indices: List[int] = []
            if token.text is None:
                raise ConfigurationError(
                    "TokenCharactersIndexer needs a tokenizer that retains text"
                )
            for character in self._character_tokenizer.tokenize(token.text):
                if getattr(character, "text_id", None) is not None:
                    # `text_id` being set on the token means that we aren't using the vocabulary, we just
                    # use this idx instead.
                    index = character.text_id
                else:
                    index = vocabulary.get_token_index(character.text, self._namespace)
                token_indices.append(index)
            indices.append(token_indices)
        return {"token_characters": indices}

    @overrides
    def get_padding_lengths(self, indexed_tokens: IndexedTokenList) -> Dict[str, int]:
        padding_lengths = {}
        padding_lengths["token_characters"] = max(
            len(indexed_tokens["token_characters"]), self._token_min_padding_length
        )
        max_num_characters = self._min_padding_length
        for token in indexed_tokens["token_characters"]:
            max_num_characters = max(len(token), max_num_characters)  # type: ignore
        padding_lengths["num_token_characters"] = max_num_characters
        return padding_lengths

    @overrides
    def as_padded_tensor_dict(
        self, tokens: IndexedTokenList, padding_lengths: Dict[str, int]
    ) -> Dict[str, torch.Tensor]:
        # Pad the tokens.
        padded_tokens = pad_sequence_to_length(
            tokens["token_characters"],
            padding_lengths["token_characters"],
            default_value=lambda: [],
        )

        # Pad the characters within the tokens.
        desired_token_length = padding_lengths["num_token_characters"]
        longest_token: List[int] = max(tokens["token_characters"], key=len, default=[])  # type: ignore
        padding_value = 0
        if desired_token_length > len(longest_token):
            # Since we want to pad to greater than the longest token, we add a
            # "dummy token" so we can take advantage of the fast implementation of itertools.zip_longest.
            padded_tokens.append([padding_value] * desired_token_length)
        # pad the list of lists to the longest sublist, appending 0's
        padded_tokens = list(zip(*itertools.zip_longest(*padded_tokens, fillvalue=padding_value)))
        if desired_token_length > len(longest_token):
            # Removes the "dummy token".
            padded_tokens.pop()
        # Truncates all the tokens to the desired length, and return the result.
        return {
            "token_characters": torch.LongTensor(
                [list(token[:desired_token_length]) for token in padded_tokens]
            )
        }

    @overrides
    def get_empty_token_list(self) -> IndexedTokenList:
        return {"token_characters": []}




from .dataset_reader import DatasetReader
from .text_classification_json_reader import TextClassificationJSONReader
from .universal_dependencies_dataset_reader import UniversalDependenciesDatasetReader
from .conllu import ConllDatasetReader



"""
Adapted from COMBO
Authors: Mateusz Klimaszewski, Lukasz Pszenny
"""
import warnings
from typing import Optional, Dict, Any

from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data import Vocabulary
from combo.data.token_indexers import IndexedTokenList
from combo.data.token_indexers.pretrained_transformer_indexer import PretrainedTransformerIndexer
from combo.data.token_indexers.pretrained_transformer_mismatched_indexer import PretrainedTransformerMismatchedIndexer


@Registry.register('pretrained_transformer_mismatched_fixed_token_indexer')
class PretrainedTransformerFixedMismatchedIndexer(PretrainedTransformerMismatchedIndexer):
    @register_arguments
    def __init__(self, model_name: str, namespace: str = "tags", max_length: int = None,
                 tokenizer_kwargs: Optional[Dict[str, Any]] = None, **kwargs) -> None:
        # The matched version v.s. mismatched
        super().__init__(model_name, namespace, max_length, tokenizer_kwargs, **kwargs)
        self._matched_indexer = PretrainedTransformerIndexer(
            model_name,
            namespace=namespace,
            max_length=max_length,
            tokenizer_kwargs=tokenizer_kwargs,
            **kwargs,
        )
        self._allennlp_tokenizer = self._matched_indexer._allennlp_tokenizer
        self._tokenizer = self._matched_indexer._tokenizer
        self._num_added_start_tokens = self._matched_indexer._num_added_start_tokens
        self._num_added_end_tokens = self._matched_indexer._num_added_end_tokens
        self._max_length = max_length or self._tokenizer.max_len_single_sentence
        if self._max_length <= 0:
            raise ValueError('Maximum sentence length must be larger than 0.')

    @overrides
    def tokens_to_indices(self,
                          tokens,
                          vocabulary: Vocabulary) -> IndexedTokenList:
        """
        Method is overridden in order to raise an error while the number of tokens needed to embed a sentence exceeds the
        maximal input of a model.
        """
        self._matched_indexer._add_encoding_to_vocabulary_if_needed(vocabulary)

        wordpieces, offsets = self._allennlp_tokenizer.intra_word_tokenize(
            [t.ensure_text() for t in tokens])

        if len(wordpieces) > self._max_length:
            warnings.warn("Following sentence consists of more wordpiece tokens that the model can process:\n" +\
                             " ".join([str(x) for x in tokens[:10]]) + " ... \n" + \
                             f"Maximal input: {self._max_length}\n"+ \
                             f"Current input: {len(wordpieces)}")

            tokens_chunk_len = self._max_length
            start_ind = 0
            wordpieces = []
            offsets = []

            while start_ind < len(tokens):
                tokens_chunk = tokens[start_ind:min(start_ind+tokens_chunk_len, len(tokens))]
                _wordpieces, _offsets = self._allennlp_tokenizer.intra_word_tokenize([t.ensure_text() for t in tokens_chunk])
                wordpieces += _wordpieces
                offset_len = offsets[-1][1] if start_ind > 0 else 0
                offsets += [(o[0]+offset_len, o[1]+offset_len) for o in _offsets]
                start_ind += tokens_chunk_len

        offsets = [x if x is not None else (-1, -1) for x in offsets]

        output: IndexedTokenList = {
            "token_ids": [t.text_id for t in wordpieces],
            "mask": [True] * len(tokens),  # for original tokens (i.e. word-level)
            "type_ids": [t.type_id for t in wordpieces],
            "offsets": offsets,
            "wordpiece_mask": [True] * len(wordpieces),  # for wordpieces (i.e. subword-level)
        }

        return self._matched_indexer._postprocess_output(output)



import errno
import logging
import os

import requests
import tqdm
import urllib3
from requests import adapters, exceptions

logger = logging.getLogger(__name__)

DATA_TO_PATH = {
    "prototype": "prototype",
    "ud213": "ud_213"
}
_URL = "http://s3.clarin-pl.eu/dspace/combo/{data}/{model}.tar.gz"
_HOME_DIR = os.getenv("HOME", os.curdir)
_CACHE_DIR = os.getenv("COMBO_DIR", os.path.join(_HOME_DIR, ".combo"))


def download_file(model_name, force=False):
    _make_cache_dir()
    split_name = model_name.split("-")
    data = split_name[-1]
    url = _URL.format(model=model_name, data=DATA_TO_PATH.get(data, data))
    local_filename = url.split("/")[-1]
    location = os.path.join(_CACHE_DIR, local_filename)
    if os.path.exists(location) and not force:
        logger.debug("Using cached model.")
        return location
    chunk_size = 1024
    logger.info(url)
    try:
        with _requests_retry_session(retries=2).get(url, stream=True) as r:
            pbar = tqdm.tqdm(unit="B", total=int(r.headers.get("content-length")),
                             unit_divisor=chunk_size, unit_scale=True)
            with open(location, "wb") as f:
                with pbar:
                    for chunk in r.iter_content(chunk_size):
                        if chunk:
                            f.write(chunk)
                            pbar.update(len(chunk))
    except exceptions.RetryError:
        raise ConnectionError(f"Couldn't find or download model {model_name}.tar.gz. "
                              "Check if model name is correct or try again later!")

    return location


def _make_cache_dir():
    try:
        os.makedirs(_CACHE_DIR)
        logger.info(f"Making cache dir {_CACHE_DIR}")
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def _requests_retry_session(
        retries=3,
        backoff_factor=0.3,
        status_forcelist=(404, 500, 502, 504),
        session=None,
):
    """Source: https://www.peterbe.com/plog/best-practice-with-retries-with-requests"""
    session = session or requests.Session()
    retry = urllib3.Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = adapters.HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/data/data_loaders/simple_data_loader.py
"""

import math
import random
from typing import Optional, List, Iterator, Callable

import torch

from combo.common.util import lazy_groups_of
from combo.common.tqdm import Tqdm
from combo.config import Registry
from combo.data.dataset_loaders.data_collator import DefaultDataCollator
from combo.data.dataset_readers import DatasetReader
from combo.data.instance import Instance
from combo.data.vocabulary import Vocabulary
import combo.nn.utils as nn_util

from combo.data.dataset_loaders.dataset_loader import DataLoader, TensorDict


@Registry.register("simple_data_loader")
@Registry.register("simple_data_loader_from_dataset_reader", "from_dataset_reader")
class SimpleDataLoader(DataLoader):
    """
    A very simple `DataLoader` that is mostly used for testing.
    """
    def __init__(
        self,
        instances: List[Instance],
        batch_size: int,
        *,
        shuffle: bool = False,
        batches_per_epoch: Optional[int] = None,
        vocabulary: Optional[Vocabulary] = None,
        collate_fn: Optional[Callable[[List[Instance]], TensorDict]] = DefaultDataCollator()
    ) -> None:
        self.instances = instances
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.batches_per_epoch = batches_per_epoch
        self.vocab = vocabulary
        self.cuda_device: Optional[torch.device] = None
        self._batch_generator: Optional[Iterator[TensorDict]] = None
        self.collate_fn = collate_fn

    def __len__(self) -> int:
        if self.batches_per_epoch is not None:
            return self.batches_per_epoch
        return math.ceil(len(self.instances) / self.batch_size)

    def __iter__(self) -> Iterator[TensorDict]:
        if self.batches_per_epoch is None:
            yield from self._iter_batches()
        else:
            if self._batch_generator is None:
                self._batch_generator = self._iter_batches()
            for i in range(self.batches_per_epoch):
                try:
                    yield next(self._batch_generator)
                except StopIteration:  # data_generator is exhausted
                    self._batch_generator = self._iter_batches()  # so refresh it
                    yield next(self._batch_generator)

    def _iter_batches(self) -> Iterator[TensorDict]:
        if self.shuffle:
            random.shuffle(self.instances)
        for batch in lazy_groups_of(self.iter_instances(), self.batch_size):
            tensor_dict = self.collate_fn(batch)
            if self.cuda_device is not None:
                tensor_dict = nn_util.move_to_device(tensor_dict, self.cuda_device)
            yield tensor_dict

    def iter_instances(self) -> Iterator[Instance]:
        for instance in self.instances:
            if self.vocab is not None:
                instance.index_fields(self.vocab)
            yield instance

    def index_with(self, vocab: Vocabulary) -> None:
        self.vocab = vocab
        for instance in self.instances:
            instance.index_fields(self.vocab)

    def set_target_device(self, device: torch.device) -> None:
        self.cuda_device = device

    @classmethod
    def from_dataset_reader(
        cls,
        reader: DatasetReader,
        data_path: str,
        batch_size: int,
        shuffle: bool = False,
        batches_per_epoch: Optional[int] = None,
        quiet: bool = False,
        collate_fn: Optional[Callable[[List[Instance]], TensorDict]] = DefaultDataCollator()
    ) -> "SimpleDataLoader":
        instance_iter = reader.read(data_path)
        if not quiet:
            instance_iter = Tqdm.tqdm(instance_iter, desc="loading instances")
        instances = list(instance_iter)
        new_obj = cls(instances, batch_size, shuffle=shuffle, batches_per_epoch=batches_per_epoch, collate_fn=collate_fn)
        new_obj.constructed_args = {
            'reader': reader,
            'data_path': data_path,
            'batch_size': batch_size,
            'shuffle': shuffle,
            'batches_per_epoch': batches_per_epoch,
            'quiet': quiet
        }
        new_obj.constructed_from = 'from_dataset_reader'
        return new_obj



"""
A default model and classes, mainly for testing the package
"""
from typing import Optional

from combo.data import DatasetReader
from combo.data.batch import Batch
from combo.data.dataset_loaders import SimpleDataLoader, DataLoader
from combo.data.dataset_readers import UniversalDependenciesDatasetReader
from combo.data.token_indexers import TokenConstPaddingCharactersIndexer, \
    TokenFeatsIndexer, SingleIdTokenIndexer, PretrainedTransformerFixedMismatchedIndexer
from combo.data.tokenizers import CharacterTokenizer
from combo.data.vocabulary import Vocabulary
from combo.combo_model import ComboModel
from combo.models.encoder import ComboEncoder, ComboStackedBidirectionalLSTM
from combo.models.dilated_cnn import DilatedCnnEncoder
from combo.modules.lemma import LemmatizerModel
from combo.modules.morpho import MorphologicalFeatures
from combo.modules.parser import DependencyRelationModel, HeadPredictionModel
from combo.modules.text_field_embedders import BasicTextFieldEmbedder
from combo.modules.token_embedders import CharacterBasedWordEmbedder, TransformersWordEmbedder
from combo.nn.activations import ReLUActivation, TanhActivation, LinearActivation
from combo.modules import FeedForwardPredictor
from combo.nn.base import Linear
from combo.nn.regularizers.regularizers import L2Regularizer
from combo.nn import RegularizerApplicator
from combo.data import Tokenizer, LamboTokenizer


def default_character_indexer(namespace=None,
                              min_padding_length: int = 32) -> TokenConstPaddingCharactersIndexer:
    if namespace:
        return TokenConstPaddingCharactersIndexer(
            tokenizer=CharacterTokenizer(end_tokens=["__END__"],
                                         start_tokens=["__START__"]),
            min_padding_length=min_padding_length,
            namespace=namespace
        )
    else:
        return TokenConstPaddingCharactersIndexer(
            tokenizer=CharacterTokenizer(end_tokens=["__END__"],
                                         start_tokens=["__START__"]),
            min_padding_length=min_padding_length
        )


def default_ud_dataset_reader(pretrained_transformer_name: str,
                              tokenizer: Optional[Tokenizer] = None) -> UniversalDependenciesDatasetReader:
    tokenizer = tokenizer or LamboTokenizer()
    return UniversalDependenciesDatasetReader(
        features=["token", "char"],
        lemma_indexers={
            "char": default_character_indexer("lemma_characters")
        },
        targets=["deprel", "head", "upostag", "lemma", "feats", "xpostag"],
        token_indexers={
            "char": default_character_indexer(),
            "feats": TokenFeatsIndexer(),
            "lemma": default_character_indexer(),
            "token": PretrainedTransformerFixedMismatchedIndexer(pretrained_transformer_name),
            "upostag": SingleIdTokenIndexer(
                feature_name="pos_",
                namespace="upostag"
            ),
            "xpostag": SingleIdTokenIndexer(
                feature_name="tag_",
                namespace="xpostag"
            )
        },
        use_sem=False,
        tokenizer=tokenizer
    )


def default_data_loader(dataset_reader: DatasetReader,
                        file_path: str,
                        batch_size: int = 16,
                        batches_per_epoch: int = 4) -> SimpleDataLoader:
    return SimpleDataLoader.from_dataset_reader(dataset_reader,
                                                data_path=file_path,
                                                batch_size=batch_size,
                                                batches_per_epoch=batches_per_epoch,
                                                shuffle=True,
                                                collate_fn=lambda instances: Batch(instances).as_tensor_dict())


def default_vocabulary(data_loader: DataLoader) -> Vocabulary:
    return Vocabulary.from_data_loader_extended(
        data_loader,
        non_padded_namespaces=['head_labels'],
        only_include_pretrained_words=False,
        oov_token='_',
        padding_token='__PAD__'
    )


def default_model(pretrained_transformer_name: str, vocabulary: Vocabulary) -> ComboModel:
    return ComboModel(
        vocabulary=vocabulary,
        dependency_relation=DependencyRelationModel(
            vocabulary=vocabulary,
            dependency_projection_layer=Linear(
                activation=TanhActivation(),
                dropout_rate=0.25,
                in_features=1024,
                out_features=128
            ),
            head_predictor=HeadPredictionModel(
                cycle_loss_n=0,
                dependency_projection_layer=Linear(
                    activation=TanhActivation(),
                    in_features=1024,
                    out_features=512
                ),
                head_projection_layer=Linear(
                    activation=TanhActivation(),
                    in_features=1024,
                    out_features=512
                )
            ),
            head_projection_layer=Linear(
                activation=TanhActivation(),
                dropout_rate=0.25,
                in_features=1024,
                out_features=128
            ),
            vocab_namespace="deprel_labels"
        ),
        lemmatizer=LemmatizerModel(
            vocabulary=vocabulary,
            activations=[ReLUActivation(), ReLUActivation(), ReLUActivation(), LinearActivation()],
            char_vocab_namespace="token_characters",
            dilation=[1, 2, 4, 1],
            embedding_dim=256,
            filters=[256, 256, 256],
            input_projection_layer=Linear(
                activation=TanhActivation(),
                dropout_rate=0.25,
                in_features=1024,
                out_features=32
            ),
            kernel_size=[3, 3, 3, 1],
            lemma_vocab_namespace="lemma_characters",
            padding=[1, 2, 4, 0],
            stride=[1, 1, 1, 1]
        ),
        loss_weights={
            "deprel": 0.8,
            "feats": 0.2,
            "head": 0.2,
            "lemma": 0.05,
            "semrel": 0.05,
            "upostag": 0.05,
            "xpostag": 0.05
        },
        morphological_feat=MorphologicalFeatures(
            vocabulary=vocabulary,
            activations=[TanhActivation(), LinearActivation()],
            dropout=[0.25, 0.],
            hidden_dims=[128],
            input_dim=1024,
            num_layers=2,
            vocab_namespace="feats_labels"
        ),
        regularizer=RegularizerApplicator([
            (".*conv1d.*", L2Regularizer(1e-6)),
            (".*forward.*", L2Regularizer(1e-6)),
            (".*backward.*", L2Regularizer(1e-6)),
            (".*char_embed.*", L2Regularizer(1e-5))
        ]),
        seq_encoder=ComboEncoder(
            layer_dropout_probability=0.33,
            stacked_bilstm=ComboStackedBidirectionalLSTM(
                hidden_size=512,
                input_size=164,
                layer_dropout_probability=0.33,
                num_layers=2,
                recurrent_dropout_probability=0.33
            )
        ),
        text_field_embedder=BasicTextFieldEmbedder(
            token_embedders={
                "char": CharacterBasedWordEmbedder(
                    vocabulary=vocabulary,
                    dilated_cnn_encoder=DilatedCnnEncoder(
                        activations=[ReLUActivation(), ReLUActivation(), LinearActivation()],
                        dilation=[1, 2, 4],
                        filters=[512, 256, 64],
                        input_dim=64,
                        kernel_size=[3, 3, 3],
                        padding=[1, 2, 4],
                        stride=[1, 1, 1],
                    ),
                    embedding_dim=64
                ),
                "token": TransformersWordEmbedder(pretrained_transformer_name, projection_dim=100)
            }
        ),
        upos_tagger=FeedForwardPredictor.from_vocab(
            vocabulary=vocabulary,
            activations=[TanhActivation(), LinearActivation()],
            dropout=[0.25, 0.],
            hidden_dims=[64],
            input_dim=1024,
            num_layers=2,
            vocab_namespace="upostag_labels"
        ),
        xpos_tagger=FeedForwardPredictor.from_vocab(
            vocabulary=vocabulary,
            activations=[TanhActivation(), LinearActivation()],
            dropout=[0.25, 0.],
            hidden_dims=[64],
            input_dim=1024,
            num_layers=2,
            vocab_namespace="xpostag_labels"
        )
    )



from typing import Tuple

from torch import optim
from combo.config import FromParameters, Registry


class Optimizer(FromParameters):
    def __init__(self):
        pass


@Registry.register("adam")
class Adam(Optimizer, optim.Adam):
    def __init__(self,
                 params,
                 lr: float = 0.001,
                 betas: Tuple[float, float] = (0.9, 0.999),
                 eps: float = 1e-08,
                 weight_decay: float = 0.):
        Optimizer.__init__(self)
        optim.Adam.__init__(self, params=params,
                            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        self.constructed_args = {
            "lr": lr,
            "betas": betas,
            "eps": eps,
            "weight_decay": weight_decay
        }



import subprocess
from setuptools import find_packages, setup

subprocess.call(['pip', 'install', '--no-deps', '--ignore-requires-python', 'git+https://gitlab.clarin-pl.eu/syntactic-tools/lambo.git#egg=lambo'])

REQUIREMENTS = [
    "absl-py~=1.4.0",
    "base58~=2.1.1",
    "cached-path~=1.3.3",
    "conllu~=4.4.1",
    "conllutils~=1.1.4",
    "dill~=0.3.6",
    "importlib-resources~=5.12.0",
    "h5py~=3.9.0",
    "overrides~=7.3.1",
    "torch~=2.0.0",
    "torchtext~=0.15.1",
    "numpy~=1.24.1",
    "pytorch-lightning~=2.0.01",
    "requests~=2.28.2",
    "tqdm~=4.64.1",
    "urllib3~=1.26.14",
    "filelock~=3.9.0",
    "pytest~=7.2.2",
    "transformers~=4.27.3",
    "spacy~=3.3.1"
]

setup(
    name="combo-nlp",
    version="3.0.1",
    author="Maja Jablonska",
    author_email="maja.jablonska@ipipan.waw.pl",
    install_requires=REQUIREMENTS,
    packages=find_packages(exclude=['tests']),
    license="GPL-3.0",
    url='https://gitlab.clarin-pl.eu/syntactic-tools/combo',
    keywords="nlp natural-language-processing dependency-parsing",
    tests_require=['pytest', 'pylint'],
    python_requires='>=3.6',
    package_data={'combo': ['config.template.json']},
    entry_points={'console_scripts': ['combo=combo.main:main']},
    classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
        ]
)


import unittest

from combo.data.vocabulary import match_namespace, Vocabulary


class VocabularyTest(unittest.TestCase):

    def test_match_namespace(self):
        self.assertTrue(match_namespace('*labels', 'test_labels'))

    def test_dont_match_namespace(self):
        self.assertFalse(match_namespace('*labels', 'test_something'))

    def test_default_non_padded_namespace(self):
        v = Vocabulary()
        self.assertFalse(v.is_padded('test_labels'))

    def test_empty_padded_namespace_is_padded(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        self.assertTrue(v.is_padded('padded_example'))

    def test_empty_non_padded_namespace_is_not_padded(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        self.assertFalse(v.is_padded('non_padded_example'))

    def test_add_token_to_padded_namespace_correct_vocab_size(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_token_to_namespace('test_token', 'padded_example')
        self.assertEqual(v.get_vocab_size('padded_example'), 3)

    def test_add_token_to_padded_namespace_correct_token_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_token_to_namespace('test_token', 'padded_example')
        self.assertEqual(v.get_token_index('test_token', 'padded_example'), 2)

    def test_add_token_to_padded_namespace_correct_token_for_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_token_to_namespace('test_token', 'padded_example')
        self.assertEqual(v.get_token_from_index(2, 'padded_example'), 'test_token')

    def test_add_tokens_to_padded_namespace_correct_vocab_size(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'padded_example')
        self.assertEqual(v.get_vocab_size('padded_example'), 4)

    def test_add_tokens_to_padded_namespace_correct_token_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'padded_example')
        self.assertEqual(v.get_token_index('test_token1', 'padded_example'), 2)
        self.assertEqual(v.get_token_index('test_token2', 'padded_example'), 3)

    def test_add_tokens_to_padded_namespace_correct_token_for_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'padded_example')
        self.assertEqual(v.get_token_from_index(2, 'padded_example'), 'test_token1')
        self.assertEqual(v.get_token_from_index(3, 'padded_example'), 'test_token2')


    def test_add_tokens_to_padded_namespace_correct_index_to_token_dict(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'padded_example')
        self.assertEqual(v.get_index_to_token_vocabulary('padded_example'),
                         {0: '@@PADDING@@',
                          1: '@@UNKNOWN@@',
                          2: 'test_token1',
                          3: 'test_token2'})

    def test_add_tokens_to_padded_namespace_correct_token_to_index_dict(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'padded_example')
        self.assertEqual(v.get_token_to_index_vocabulary('padded_example'),
                         {'@@PADDING@@': 0,
                          '@@UNKNOWN@@': 1,
                          'test_token1': 2,
                          'test_token2': 3})

    def test_add_token_to_non_padded_namespace_correct_vocab_size(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_token_to_namespace('test_token', 'non_padded_example')
        self.assertEqual(v.get_vocab_size('non_padded_example'), 1)

    def test_add_token_to_non_padded_namespace_correct_token_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_token_to_namespace('test_token', 'non_padded_example')
        self.assertEqual(v.get_token_index('test_token', 'non_padded_example'), 0)

    def test_add_token_to_non_padded_namespace_correct_token_for_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_token_to_namespace('test_token', 'non_padded_example')
        self.assertEqual(v.get_token_from_index(0, 'non_padded_example'), 'test_token')

    def test_add_tokens_to_non_padded_namespace_correct_vocab_size(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'non_padded_example')
        self.assertEqual(v.get_vocab_size('non_padded_example'), 2)

    def test_add_tokens_to_non_padded_namespace_correct_token_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'non_padded_example')
        self.assertEqual(v.get_token_index('test_token1', 'non_padded_example'), 0)
        self.assertEqual(v.get_token_index('test_token2', 'non_padded_example'), 1)

    def test_add_tokens_to_non_padded_namespace_correct_token_for_index(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'non_padded_example')
        self.assertEqual(v.get_token_from_index(0, 'non_padded_example'), 'test_token1')
        self.assertEqual(v.get_token_from_index(1, 'non_padded_example'), 'test_token2')

    def test_add_tokens_to_non_padded_namespace_correct_index_to_token_dict(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'non_padded_example')
        self.assertEqual(v.get_index_to_token_vocabulary('non_padded_example'),
                         {0: 'test_token1',
                          1: 'test_token2'})

    def test_add_tokens_to_non_padded_namespace_correct_token_to_index_dict(self):
        padded_namespaces = ['padded_example']
        non_padded_namespaces = ['non_padded_example']
        v = Vocabulary(non_padded_namespaces=non_padded_namespaces)
        v.add_tokens_to_namespace(['test_token1', 'test_token2'], 'non_padded_example')
        self.assertEqual(v.get_token_to_index_vocabulary('non_padded_example'),
                         {'test_token1': 0,
                          'test_token2': 1})


if __name__ == '__main__':
    unittest.main()



from typing import Optional

import torch
from overrides import overrides

from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data.vocabulary import Vocabulary
from combo.modules.token_embedders.token_embedder import Embedding


@Registry.register("projected_words_embedder")
class ProjectedWordEmbedder(Embedding):
    """Word embeddings."""
    @register_arguments
    def __init__(self,
                 embedding_dim: int,
                 num_embeddings: int = None,
                 weight: torch.FloatTensor = None,
                 padding_index: int = None,
                 trainable: bool = True,
                 max_norm: float = None,
                 norm_type: float = 2.0,
                 scale_grad_by_freq: bool = False,
                 sparse: bool = False,
                 vocab_namespace: str = "tokens",
                 pretrained_file: str = None,
                 vocabulary: Vocabulary = None,
                 projection_layer: Optional["Linear"] = None): # Change to Optional[Linear]
        super().__init__(
            embedding_dim=embedding_dim,
            num_embeddings=num_embeddings,
            weight=weight,
            padding_index=padding_index,
            trainable=trainable,
            max_norm=max_norm,
            norm_type=norm_type,
            scale_grad_by_freq=scale_grad_by_freq,
            sparse=sparse,
            vocab_namespace=vocab_namespace,
            pretrained_file=pretrained_file,
            vocabulary=vocabulary
        )
        self._projection = projection_layer
        self.output_dim = embedding_dim if projection_layer is None else projection_layer.out_features

    @overrides
    def get_output_dim(self) -> int:
        return self.output_dim



"""
Partially adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/data/dataset_readers/text_classification_json.py
"""

import json
from typing import Dict, Iterable, Optional, Union, List

from overrides import overrides

from combo.data.dataset_readers.dataset_reader import DatasetReader, DatasetReaderInput
from combo.data.dataset_readers.utils import MalformedFileException
from combo.data import Instance, Tokenizer, TokenIndexer
from combo.data.fields import Field, ListField
from combo.data.fields.label_field import LabelField
from combo.data.fields.text_field import TextField
from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.utils import ConfigurationError


def _is_sentence_segmenter(sentence_segmenter: Optional[Tokenizer]) -> bool:
    split_sentences_method = getattr(sentence_segmenter, "split_sentences", None)
    return callable(split_sentences_method)


@Registry.register('text_classification_json_dataset_reader')
class TextClassificationJSONReader(DatasetReader):
    @register_arguments
    def __init__(self,
                 tokenizer: Optional[Tokenizer] = None,
                 token_indexers: Optional[Dict[str, TokenIndexer]] = None,
                 sentence_segmenter: Optional[Tokenizer] = None,
                 max_sequence_length: Optional[int] = None,
                 skip_label_indexing: bool = False,
                 text_key: str = "text",
                 label_key: str = "label") -> None:
        if ((sentence_segmenter is not None) and
                (not _is_sentence_segmenter(sentence_segmenter))):
            raise ConfigurationError(f'Passed sentence segmenter has no'
                                     f'split_sentences method!')

        super().__init__(tokenizer, token_indexers)
        self.__sentence_segmenter = sentence_segmenter
        self.__max_sequence_length = max_sequence_length
        self.__skip_label_indexing = skip_label_indexing
        self.__text_key = text_key
        self.__label_key = label_key

    @property
    def sentence_segmenter(self) -> Optional[Tokenizer]:
        return self.__sentence_segmenter

    @sentence_segmenter.setter
    def sentence_segmenter(self, value: Optional[Tokenizer]):
        if ((value is not None) and
                (not _is_sentence_segmenter(value))):
            raise ConfigurationError(f'Passed sentence segmenter has no'
                                     f'split_sentences method!')
        self.__sentence_segmenter = value

    @property
    def max_sequence_length(self) -> Optional[int]:
        return self.__max_sequence_length

    @max_sequence_length.setter
    def max_sequence_length(self, value: Optional[int]):
        self.__max_sequence_length = value

    @property
    def skip_label_indexing(self) -> bool:
        return self.__skip_label_indexing

    @skip_label_indexing.setter
    def skip_label_indexing(self, value: bool):
        self.__skip_label_indexing = value

    @property
    def text_key(self) -> str:
        return self.__text_key

    @text_key.setter
    def text_key(self, value: str):
        self.__text_key = value

    @property
    def label_key(self) -> str:
        return self.__label_key

    @label_key.setter
    def label_key(self, value: str):
        self.__label_key = value

    @overrides
    def _read(self, file_path: DatasetReaderInput) -> Iterable[Instance]:
        # if self.file_path is None:
        #     raise ValueError('File path is None')
        with open(file_path, "r") as data_file:
            for line in data_file.readlines():
                if not line:
                    continue
                items = json.loads(line)
                text = items.get(self.text_key)
                if text is None:
                    raise MalformedFileException(f'No item with {self.text_key} (text) label')
                label = items.get(self.label_key)
                if label is not None:
                    if self.skip_label_indexing:
                        try:
                            label = int(label)
                        except ValueError:
                            raise MalformedFileException("Labels must be integers if skip_label_indexing is True.")
                    else:
                        label = str(label)
                yield self.text_to_instance(text, label)


    def text_to_instance(self,
                         text: str,
                         label: Optional[Union[str, int]] = None) -> Instance:
        """

        :param text: the text to classify
        :param label: the label for the text
        :return: Instance containing the following fields:
            - tokens ('TextField')
            - label ('LabelField')
        """
        fields: Dict[str, Field] = {}
        if self.sentence_segmenter is not None:
            sentences: List[Field] = []

            # TODO: some subclass for sentence segmenter for tokenizers?
            sentence_splits = self.sentence_segmenter.split_sentences(text)
            for sentence in sentence_splits:
                word_tokens = self.tokenizer.tokenize(sentence)
                if self.max_sequence_length is not None:
                    word_tokens = self._truncate(word_tokens)
                sentences.append(TextField(word_tokens))
            fields["tokens"] = ListField(sentences)
        else:
            tokens = self.tokenizer.tokenize(text)
            if self.max_sequence_length is not None:
                tokens = self._truncate(tokens)
            fields["tokens"] = TextField(tokens)

        if label is not None:
            fields["label"] = LabelField(label,
                                         skip_indexing=self.skip_label_indexing)
        return Instance(fields)

    def _truncate(self, tokens):
        """
        truncate a set of tokens using the provided sequence length
        """
        if len(tokens) > self.max_sequence_length:
            tokens = tokens[: self.max_sequence_length]
        return tokens

    def __call__(self, file_path: str):
        self.file_path = file_path
        return self

    @overrides
    def apply_token_indexers(self, instance: Instance) -> None:
        if self.sentence_segmenter is not None:
            for text_field in instance.fields["tokens"]:  # type: ignore
                text_field._token_indexers = self.token_indexers

        else:
            instance.fields["tokens"]._token_indexers = self.token_indexers  # type: ignore



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/data/tokenizers/token_class.py
"""
from typing import Any, Dict, List, Optional, Tuple, Union
import logging
from dataclasses import dataclass, field
import conllu

logger = logging.getLogger(__name__)


def _assert_none_or_type(value: Any, type_to_check: type) -> bool:
    return value is None or isinstance(value, type_to_check)


@dataclass(init=False, repr=False)
class Token:
    __slots__ = [
        "text",
        "idx",
        "lemma",
        "upostag",
        "xpostag",
        "entity_type",
        "feats",
        "head",
        "deprel",
        "deps",
        "misc",
        "subwords",
        "semrel",
        "embeddings",
        "text_id",
        "type_id"
    ]

    text: Optional[str]
    idx: Optional[Union[int, Tuple]]
    lemma: Optional[str]
    upostag: Optional[str]  # Coarse-grained part of speech? pos_?
    xpostag: Optional[str]  # Fine-grained part of speech? tag_ ?
    entity_type: Optional[str]
    feats: Optional[str]
    head: Optional[int]
    deprel: Optional[str]  # dep_ ?
    deps: Optional[str]
    misc: Optional[str]
    subwords: Optional[List[str]]
    semrel: Optional[str]
    embeddings: Dict[str, List[float]]
    text_id: Optional[int]
    type_id: Optional[int]

    def __init__(self,
                 text: str = None,
                 idx: Union[int, Tuple] = None,
                 lemma: str = None,
                 upostag: str = None,
                 xpostag: str = None,
                 entity_type: str = None,
                 feats: Optional[Dict[str, str]] = None,
                 head: int = None,
                 deprel: str = None,
                 deps: str = None,
                 misc: str = None,
                 subwords: List[str] = None,
                 semrel: str = None,
                 embeddings: Dict[str, List[float]] = None,
                 text_id: int = None,
                 type_id: int = None,) -> None:
        _assert_none_or_type(text, str)

        self.text = text
        self.idx = idx
        self.lemma = lemma
        self.upostag = upostag
        self.xpostag = xpostag
        self.entity_type = entity_type
        self.feats = feats
        self.head = head
        self.deprel = deprel
        self.deps = deps
        self.misc = misc
        self.subwords = subwords if subwords else []
        self.semrel = semrel

        if embeddings is None:
            # what?
            self.embeddings = field(default_factory=list, repr=False)
        else:
            self.embeddings = embeddings

        self.text_id = text_id
        self.type_id = type_id

    def as_dict(self, semrel: bool = True) -> Dict[str, Any]:
        repr = {}
        repr_keys = [
            'text', 'idx', 'lemma', 'upostag', 'xpostag', 'entity_type', 'feats',
            'head', 'deprel', 'deps', 'misc'
        ]
        for rk in repr_keys:
            repr[rk] = self.__getattribute__(rk)

        if semrel:
            repr['semrel'] = self.semrel

        return repr

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.__str__()

    def ensure_text(self) -> str:
        """
        Return the `text` field, raising an exception if it's `None`.
        """
        if self.text is None:
            raise ValueError("Unexpected null text for token")
        else:
            return self.text

    def get(self, key, default=None):
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return default

    def __getitem__(self, key):
        return getattr(self, key)

    def __show__(self):
        return (
            f"{self.text} "
            f"(idx: {self.idx}) "
            f"(lemma: {self.lemma}) "
            f"(upostag: {self.upostag}) "
            f"(xpostag: {self.xpostag}) "
            f"(entity_type: {self.entity_type}) "
            f"(feats: {self.feats}) "
            f"(head: {self.head}) "
            f"(deprel: {self.deprel}) "
            f"(deps: {self.deps}) "
            f"(misc: {self.misc}) "
            f"(subwords: {','.join(self.subwords)})"
            f"(semrel: {self.semrel}) "
            f"(embeddings: {self.embeddings}) "
            f"(text_id: {self.text_id}) "
            f"(type_id: {self.type_id}) "
        )

    @classmethod
    def from_conllu_token(cls,
                          conllu_token: conllu.models.Token):
        return cls(text=conllu_token.get("token"),
                   idx=conllu_token.get("id"),
                   lemma=conllu_token.get("lemma"),
                   upostag=conllu_token.get("upos"),
                   xpostag=conllu_token.get("xpos"),
                   feats=conllu_token.get("feats"),
                   head=conllu_token.get("head"),
                   deprel=conllu_token.get("deprel"),
                   deps=conllu_token.get("deps"),
                   misc=conllu_token.get("misc"))
